#include "LBM_3D.h"


//Shader velocity_shader;

  double time_copy_to_cpu_1 = 0;
  double time_copy_to_cpu_2 = 0;

  real_t *mpi_send_buff;

LBM_3D::LBM_3D(int Id, uchar *data, real_t omegaInit, int sizeX, int sizeY, int sizeZ, int ext, 
               int ghostLayersNorth, int ghostLayersSouth, bool writeToFile, CommandQueue inputQueue, real_t forceInit)
{

  index = Id;
  queue = inputQueue;

  memoryUsageCPU = 0;
  memoryUsageGPU = 0;

  north_neighbour = NULL;
  south_neighbour = NULL;
  north_rank = -1;
  south_rank = -1;

  ghost_layers_north = ghostLayersNorth;
  ghost_layers_south = ghostLayersSouth;

  write_to_file = writeToFile;

  // The lattice size
  size_x = sizeX;
  size_y = sizeY;
  size_z = sizeZ;

  // Relaxation parameter
  omega = omegaInit;


  // Reset the total simulation time to zero
  total_lbm_time = 0;
  total_lbm_wall_time = 0;
  time_collide = 0;
  time_collide_inner = 0.0;
  time_collide_ghost = 0.0;

  time_stream = 0;
  time_collide_last_itter = 0;
  time_stream_last_itter = 0;

  time_lbm = 0;
  time_lines = 0;
  time_avg_velocity = 0;
  time_copy_to_cpu = 0;




  time_copy_to_gpu = 0;
  time_exchange = 0;
  time_convergence_check = 0;
  time_copy_to_cpu_convergence = 0;


  // The maximum velocity obtained in simulation step
  u_max = 0;



  // The best mlups in simulation
  best_mlups = 0.;

  // Extra lattice locations	
  extra = ext;

  //hits = 0;

  viscosity = (1./6. * (2./omega - 1.));
  force = forceInit;



  //Create the Host data structure
  // Allocate host memory for solids
  solid_h = (int*) malloc(getSizeX()*getSizeY()*(getSizeZ() + ghostLayersNorth + ghostLayersSouth)*sizeof(int)); 
  if(solid_h == NULL){
    printf("unable to allocate memory for solid_h\n");
    PAUSE
  }
#ifdef DEBUG
  memoryUsageCPU += getSizeX()*getSizeY()*(getSizeZ() + ghostLayersNorth + ghostLayersSouth)*sizeof(int);

#endif

  createHostDataStructureAndCalculateTotalAndActiveVoxels(data, ghostLayersNorth, ghostLayersSouth);
  calculateGhostElemetsNorthAndSouth(ghostLayersNorth,ghostLayersSouth);

  
/*
  for(int z=0; z<size_z + ghostLayersNorth + ghostLayersSouth; z++)
  {
  printf("layer %d \n",z);
  for(int y=0; y<size_y; y++)	
  {
  for(int x=0; x<size_x; x++)
  {

  //if(solid_h[(z * size_y * size_x) + (y * size_x) + x] == -1){
  //  printf("2 ");  
  //}else{
  //  printf("0 ");  
  //}

  //printf("%d ", solid_h[(z * size_y * size_x) + (y * size_x) + x]);  
  printf("%d ", data[(z * size_y * size_x) + (y * size_x) + x]);  
  }
  printf("\n");
  }
  printf("\n");
  }

  
 exit(0);
*/

  //printf("Node: %d nl: %d nse %d nre: %d    sl: %d ses: %d sre: %d\n",index,ghost_layers_north,ghost_element_send_north,ghost_element_recive_north,ghost_layers_south,ghost_element_send_south,ghost_element_recive_south);

  try{

    //allocat memory for the ghost layers, 
    if (ghost_element_send_north != 0){
      north_send_buffer = (real_t *) malloc(ghost_element_send_north * N_VEC * sizeof(real_t));
      north_recv_buffer = (real_t *) malloc(ghost_element_recive_north * N_VEC * sizeof(real_t));

      d_north_send_buffer = createBuffer(ghost_element_send_north * N_VEC * sizeof(real_t),CL_MEM_READ_WRITE);
      d_north_recv_buffer = createBuffer(ghost_element_recive_north * N_VEC * sizeof(real_t),CL_MEM_READ_WRITE);


#ifdef DEBUG
      memoryUsageCPU += ghost_element_send_north * N_VEC * sizeof(real_t);
      memoryUsageCPU += ghost_element_recive_north * N_VEC * sizeof(real_t);

      memoryUsageGPU += ghost_element_send_north * N_VEC * sizeof(real_t);
      memoryUsageGPU += ghost_element_recive_north * N_VEC * sizeof(real_t);
#endif 
    }

    if(ghost_element_send_south != 0 ){
      south_send_buffer = (real_t *) malloc(ghost_element_send_south * N_VEC * sizeof(real_t));
      south_recv_buffer = (real_t *) malloc(ghost_element_recive_south * N_VEC * sizeof(real_t));

      d_south_send_buffer = createBuffer(ghost_element_send_south * N_VEC * sizeof(real_t),CL_MEM_READ_WRITE);
      d_south_recv_buffer = createBuffer(ghost_element_recive_south * N_VEC * sizeof(real_t),CL_MEM_READ_WRITE);




#ifdef DEBUG
      memoryUsageCPU += ghost_element_send_south * N_VEC * sizeof(real_t);
      memoryUsageCPU += ghost_element_recive_south * N_VEC * sizeof(real_t);

      memoryUsageGPU += ghost_element_send_south * N_VEC * sizeof(real_t);
      memoryUsageGPU += ghost_element_recive_south * N_VEC * sizeof(real_t);
#endif
    }



    //Kernels
    colide_and_swap          = createKernel("collide_and_swap");
    stream_and_swap          = createKernel("stream_and_swap");
    avrage_velocity          = createKernel("avrage_velocity_ux");
    reduction                = createKernel("reduce");
    copy_to_from_ghostBuffer = createKernel("copy_to_from_ghost_layer");


  } catch(Error error) {
    std::cout << "LBM_3D " << index  << " constructor" << std::endl;
    std::cout << error.what() << "(" << error.err() << ")" << std::endl;
    std::cout << getCLErrorString(error.err()) << std::endl;
    
    PAUSE
    exit(-1);
  }


  // Allocate host memory for velocity fields
  u_new = (real_t *) malloc(total_voxels*sizeof(real_t));

  if(u_new == NULL){
    printf("unable to allocate memory for u_new\n");
    system("pause");
  }
  memset(u_new,0,total_voxels*sizeof(real_t));

  max_movments = (real_t *) malloc(total_voxels*sizeof(real_t));

  if(max_movments == NULL){
    printf("unable to allocate memory for max movment\n");
    system("pause");
  }
  memset(max_movments ,0,total_voxels*sizeof(real_t));


  allocateMemoryGPU(getSizeX(), getSizeY(), getSizeZ()+ghostLayersNorth + ghostLayersSouth,
                    total_voxels, solid_h,d_f,d_avrage_velocity,d_solids,time_allocate_memory);
#if DEBUG
  memoryUsageGPU += ((N_VEC + 1 ) * total_voxels) * sizeof(real_t); // d_f's and d_avrage_velocity
  memoryUsageGPU += (getSizeX() * getSizeY() * (getSizeZ()+ghostLayersNorth + ghostLayersSouth)) * sizeof(int); // d_solid IMAGE 3D

  memoryUsageCPU += total_voxels*sizeof(real_t); //u_new
  memoryUsageCPU += total_voxels*sizeof(real_t); //max_movment
#endif



  // Open simulation log file
  std::string logfile_n;
  logfile_n += LOGFOLDER;
  logfile_n += "LBM";

  stringstream out_index;
  out_index << index;
  logfile_n += "_" + out_index.str();

  logfile_n += ".txt";

  pFile = fopen(logfile_n.c_str() ,"a+");


  createFileName();
  free(solid_h);

#if DEBUG
  printInitalState();
#endif
}


LBM_3D::~LBM_3D()
{
  freeMemoryGPU();

  // Free host memory for velocity fields
  free(u_new);

  free(max_movments);
  // Free host memory for solid
  //free(solid_h);



  // Close simulation log file
  fclose(pFile);


  //stream_and_swap.release();
  //colide_and_swap.release();
  //avrage_velocity.release();

  //todo: MPi buffering
  
  /*
  int size;
  MPI_Buffer_detach(&mpi_send_buff, &size);
  free(mpi_send_buff);
  */

}

int LBM_3D::getSizeX()
{
  return	size_x;
}

int LBM_3D::getSizeY()
{
  return	size_y;
}

int LBM_3D::getSizeZ()
{
  return	size_z;
}

void LBM_3D::createHostDataStructureAndCalculateTotalAndActiveVoxels(uchar *data, int ghostLayersNorth, int ghostLayersSouth)
{

  int zEnd =  getSizeZ()+ghostLayersNorth + ghostLayersSouth;
  int voxelIndex = 0;
  int index = 0;
  int obsIndex = 0;
  total_voxels = 0;
  voxelsActive = 0;

  for(int z= 0 ; z< zEnd; z++)
  {
    for(int y=0; y<getSizeY(); y++)
    {
      for(int x=0; x<getSizeX(); x++)
      {	

        if(data[index] == 0){
          obsIndex = voxelIndex++;

          //The active coxels is only the ones insid the actual stone loaded
          if (x >= (extra/2) && x < (getSizeX() - (extra/2))){
            if (z >= ghostLayersNorth && z < ghostLayersNorth + getSizeZ()){
              voxelsActive++; 
            }
          }

          //Total voxel is all foid voxels insid the active area including the extra padding in x direction and ghost layer
          total_voxels++;
        }else{
          obsIndex = -1;
        }
        solid_h[index] = obsIndex;
        index++;    
      }
    }
  }

}

void LBM_3D::calculateGhostElemetsNorthAndSouth(int ghostLayersNorth, int ghostLayersSouth)
{
  ghost_element_recive_north = 0;
  ghost_element_recive_south = 0;

  ghost_element_send_north = 0;
  ghost_element_send_south = 0;

  // Ant elements recived north
  for(int z = 0; z< ghostLayersNorth ; z++){
    for(int y = 0; y< getSizeY(); y++){
      for(int x = 0; x< getSizeX(); x++){
        if(getSolid(x,y,z) != -1){
          ghost_element_recive_north++;
        }
      }
    }
  }

  // Ant elements send north
  for(int z = ghostLayersNorth; z< 2*ghostLayersNorth ; z++){
    for(int y = 0; y< getSizeY(); y++){
      for(int x = 0; x< getSizeX(); x++){
        if(getSolid(x,y,z) != -1){
          ghost_element_send_north++;
        }
      }
    }
  }

  // Ant elements recived south
  for(int z = ghostLayersNorth + getSizeZ(); z< ghostLayersNorth + getSizeZ() +  ghostLayersSouth; z++){
    for(int y = 0; y< getSizeY(); y++){
      for(int x = 0; x< getSizeX(); x++){
        if(getSolid(x,y,z) != -1){
          ghost_element_recive_south++;
        }
      }
    }
  }

  // Ant elements send south
  for(int z = ghostLayersNorth + getSizeZ() - ghostLayersSouth ; z< ghostLayersNorth + getSizeZ(); z++){
    for(int y = 0; y< getSizeY(); y++){
      for(int x = 0; x< getSizeX(); x++){
        if(getSolid(x,y,z) != -1){
          ghost_element_send_south++;
        }
      }
    }
  }

}



int LBM_3D::getIndex(int x, int y, int z)
{
  return x + y*getSizeX() + z*getSizeX()*getSizeY();
}

int LBM_3D::getSolid(int x, int y, int z)
{
  int solid = solid_h[getIndex(x, y, z)];
  return solid;
}

void LBM_3D::doCollide(){
  collideAndSwapGPU(queue, colide_and_swap, 
	  getSizeX(),getSizeY(),getSizeZ()+ghost_layers_north+ghost_layers_south,
	  0,0,0,
	  omega, force, &collide_events[0],d_f,d_solids,index );
}

void LBM_3D::doCollide_inner(){
	collideAndSwapGPU(queue, colide_and_swap, 
		getSizeX(),getSizeY(),getSizeZ(), 
		0,0,ghost_layers_north,
		omega, force, &collide_events[0],d_f,d_solids,index );
}
void LBM_3D::doCollide_ghost(){
	//North gost layer
	if(ghost_layers_north != 0){
		collideAndSwapGPU(queue, colide_and_swap, 
			getSizeX(),getSizeY(),ghost_layers_north, 
			0,0,0,
			omega, force, &collide_events[1],d_f,d_solids,index );
	}

	//South gost layer
	if(ghost_layers_south != 0){
		collideAndSwapGPU(queue, colide_and_swap, 
			getSizeX(),getSizeY(),ghost_layers_south, 
			0,0,ghost_layers_north + getSizeZ(),
			omega, force, &collide_events[2],d_f,d_solids,index );
	}
}



//real_t prev_max = 0.0;
//int it = 0;
void LBM_3D::doStream(){
  streamAndSwapGPU(queue, stream_and_swap, getSizeX(),getSizeY(),getSizeZ()+ghost_layers_north+ghost_layers_south, &stream_event,d_f,d_solids,d_avrage_velocity);
  }

void LBM_3D::getMaxParticleMovment(real_t *max_movment){
  long t1, t2;
  Event event;
  *max_movment = 0.0;
  readBuffer(queue,d_avrage_velocity,max_movments,total_voxels*sizeof(real_t),&event);
  queue.finish();

  time_copy_to_cpu_convergence += executionTime(event);

  getTimeInMs(t1);
  for (int i = 0; i < total_voxels; i++)
  {
    *max_movment = max(*max_movment,max_movments[i]);
  }
  getTimeInMs(t2);

  time_convergence_check += (double (t2 - t1) / toSec);

}

void LBM_3D::calculateAvrageVelocity(real_t *local_avg_vel){

  for (int i = 0; i < total_voxels; i++)
  {
    max_movments[i] = 0.0;
  }
  writeBuffer(queue,d_avrage_velocity,max_movments,total_voxels * sizeof(real_t),NULL);
  queue.finish();

  getAverageVelocityXGPU(queue, avrage_velocity,reduction,d_f, d_avrage_velocity, d_solids,
    total_voxels, getSizeX(), getSizeY(), getSizeZ(),extra,avrage_events, 
    ghost_layers_north, ghost_layers_south, local_avg_vel);
}



void LBM_3D::exhangeGhostLayers_threaded(){

  long t1;
  long t2;

  //Start timer
  getTimeInMs(t1);
  //exhange north
  if (north_neighbour != NULL){
    memcpy(north_recv_buffer,north_neighbour->getSouthSendBuffer(),ghost_element_recive_north * N_VEC * sizeof(real_t));
  }
  //exhange south
  if (south_neighbour != NULL){
    memcpy(south_recv_buffer,south_neighbour->getNorthSendBuffer(),ghost_element_recive_south * N_VEC * sizeof(real_t));
  }
  //stop timer
  getTimeInMs(t2);
  timer = (double (t2 - t1) / toSec);
  time_exchange += timer;
}

void LBM_3D::exhangeGhostLayers_MPI(){
#if MODE == mode_mpi || MODE == mode_hybrid
  
  long t1;
  long t2;
  //MPI_Request *requests_north = new MPI_Request[2];
  //MPI_Status *statuss_north = new MPI_Status[2];

  
  //MPI_Request *requests_south = new MPI_Request[2];
  //MPI_Status *statuss_south = new MPI_Status[2];

  //Start timer
#if MODE == mode_mpi //can not have barrierer when in hybrid mode, dont know why
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  getTimeInMs(t1);

  if (ghost_layers_north != 0 && north_rank != -1){
    //send north
    //printf("%d sending  %d to %d \n",index,ghost_element_send_north,north_rank);
    //MPI_Isend(north_send_buffer,ghost_element_send_north* N_VEC,MPI_FLOAT,north_rank,0,MPI_COMM_WORLD,&requests_north[0]);
    MPI_Start(&send_north);
    
    //MPI_Isend(north_send_buffer,1,MPI_FLOAT,north_rank,0,MPI_COMM_WORLD,&requests_north[0]);
    
    //recive north
    //printf("%d recv  %d from %d \n",index,ghost_element_recive_north,north_rank);
    //MPI_Irecv(north_recv_buffer,ghost_element_recive_north* N_VEC,MPI_FLOAT,north_rank,0,MPI_COMM_WORLD,&requests_north[1]);
    MPI_Start(&reciv_north);
    
    //MPI_Irecv(north_recv_buffer,1,MPI_FLOAT,north_rank,0,MPI_COMM_WORLD,&requests_north[1]);
  }

  if(ghost_layers_south != 0 && south_rank != -1){
    //send south
    //printf("%d sending  %d to %d \n",index,ghost_element_send_south,south_rank);
    //MPI_Isend(south_send_buffer,ghost_element_send_south* N_VEC,MPI_FLOAT,south_rank,0,MPI_COMM_WORLD,&requests_south[0]);
    MPI_Start(&send_south);
    
    //MPI_Isend(south_send_buffer,1,MPI_FLOAT,south_rank,0,MPI_COMM_WORLD,&requests_south[0]);
    
    //recive south
    //printf("%d recv  %d from %d \n",index,ghost_element_recive_south,south_rank);
    //MPI_Irecv(south_recv_buffer,ghost_element_recive_south* N_VEC,MPI_FLOAT,south_rank,0,MPI_COMM_WORLD,&requests_south[1]);
    MPI_Start(&reciv_south);
    
    //MPI_Irecv(south_recv_buffer,1,MPI_FLOAT,south_rank,0,MPI_COMM_WORLD,&requests_south[1]);
  }

  if (ghost_layers_north != 0 && north_rank != -1){
    //MPI_Waitall(2, requests_north, statuss_north);
    MPI_Wait(&send_north,MPI_STATUSES_IGNORE);
    MPI_Wait(&reciv_north,MPI_STATUSES_IGNORE);
  }

  if(ghost_layers_south != 0 && south_rank != -1){
    //MPI_Waitall(2, requests_south, statuss_south);
    MPI_Wait(&send_south,MPI_STATUSES_IGNORE);
    MPI_Wait(&reciv_south,MPI_STATUSES_IGNORE);
  }
//#if MODE == mode_mpi //can not have barrierer when in hybrid mode, dont know why
//  MPI_Barrier(MPI_COMM_WORLD);
//#endif
  //stop timer
  getTimeInMs(t2);
  timer = (double (t2 - t1) / toSec);
  time_exchange += timer;
#endif
}


void LBM_3D::copyGhostLayersToGPU(){

  try{

  copy_to_from_ghostBuffer.setArg(0,d_f[0]);
  copy_to_from_ghostBuffer.setArg(1,d_f[1]);
  copy_to_from_ghostBuffer.setArg(2,d_f[2]);
  copy_to_from_ghostBuffer.setArg(3,d_f[3]);
  copy_to_from_ghostBuffer.setArg(4,d_f[4]);
  copy_to_from_ghostBuffer.setArg(5,d_f[5]);
  copy_to_from_ghostBuffer.setArg(6,d_f[6]);
  copy_to_from_ghostBuffer.setArg(7,d_f[7]);
  copy_to_from_ghostBuffer.setArg(8,d_f[8]);
  copy_to_from_ghostBuffer.setArg(9,d_f[9]);
  copy_to_from_ghostBuffer.setArg(10,d_f[10]);
  copy_to_from_ghostBuffer.setArg(11,d_f[11]);
  copy_to_from_ghostBuffer.setArg(12,d_f[12]);
  copy_to_from_ghostBuffer.setArg(13,d_f[13]);
  copy_to_from_ghostBuffer.setArg(14,d_f[14]);
  copy_to_from_ghostBuffer.setArg(15,d_f[15]);
  copy_to_from_ghostBuffer.setArg(16,d_f[16]);
  copy_to_from_ghostBuffer.setArg(17,d_f[17]);
  copy_to_from_ghostBuffer.setArg(18,d_f[18]);

  copy_to_from_ghostBuffer.setArg(20,N_VEC);




  //Coppying north layer
if (ghost_layers_north != 0){
  
  //Write the host data to the buffer
  writeBuffer(queue, d_north_recv_buffer,north_recv_buffer,ghost_element_recive_north *N_VEC * sizeof(real_t),&copy_to_device_events[0]);
  
  copy_to_from_ghostBuffer.setArg(19,d_north_recv_buffer); //Buffer
  copy_to_from_ghostBuffer.setArg(21,0); //Offset startindex
  copy_to_from_ghostBuffer.setArg(22,0); //Direction


  NDRange global(ghost_element_recive_north);
  queue.enqueueNDRangeKernel(copy_to_from_ghostBuffer,NullRange,global,NullRange,NULL,&copy_to_device_events[1]);

}
  //Coppying south layer
if (ghost_layers_south != 0){

  //Write the host data to the buffer
  writeBuffer(queue, d_south_recv_buffer,south_recv_buffer,ghost_element_recive_south * N_VEC * sizeof(real_t),&copy_to_device_events[2]);
  
  copy_to_from_ghostBuffer.setArg(19,d_south_recv_buffer); //Buffer
  copy_to_from_ghostBuffer.setArg(21,(total_voxels - ghost_element_recive_south)); //Offset startindex
  copy_to_from_ghostBuffer.setArg(22,0); //Direction


  NDRange global(ghost_element_recive_south);
  queue.enqueueNDRangeKernel(copy_to_from_ghostBuffer,NullRange,global,NullRange,NULL,&copy_to_device_events[3]);

}

  } catch(Error error) {
    std::cout << "LBM " << index << " copy to gpu" << std::endl;
   std::cout << error.what() << "(" << error.err() << ")" << std::endl;
   std::cout << getCLErrorString(error.err()) << std::endl;
   system("pause");
   exit(-1);
  }

}
void LBM_3D::copyGhostLayersToCPU(){
  try{

    copy_to_from_ghostBuffer.setArg(0,d_f[0]);
    copy_to_from_ghostBuffer.setArg(1,d_f[1]);
    copy_to_from_ghostBuffer.setArg(2,d_f[2]);
    copy_to_from_ghostBuffer.setArg(3,d_f[3]);
    copy_to_from_ghostBuffer.setArg(4,d_f[4]);
    copy_to_from_ghostBuffer.setArg(5,d_f[5]);
    copy_to_from_ghostBuffer.setArg(6,d_f[6]);
    copy_to_from_ghostBuffer.setArg(7,d_f[7]);
    copy_to_from_ghostBuffer.setArg(8,d_f[8]);
    copy_to_from_ghostBuffer.setArg(9,d_f[9]);
    copy_to_from_ghostBuffer.setArg(10,d_f[10]);
    copy_to_from_ghostBuffer.setArg(11,d_f[11]);
    copy_to_from_ghostBuffer.setArg(12,d_f[12]);
    copy_to_from_ghostBuffer.setArg(13,d_f[13]);
    copy_to_from_ghostBuffer.setArg(14,d_f[14]);
    copy_to_from_ghostBuffer.setArg(15,d_f[15]);
    copy_to_from_ghostBuffer.setArg(16,d_f[16]);
    copy_to_from_ghostBuffer.setArg(17,d_f[17]);
    copy_to_from_ghostBuffer.setArg(18,d_f[18]);

    copy_to_from_ghostBuffer.setArg(20,N_VEC);

    //Coppying north layer
    if (ghost_layers_north != 0){


      copy_to_from_ghostBuffer.setArg(19,d_north_send_buffer); //Buffer
      copy_to_from_ghostBuffer.setArg(21,ghost_element_recive_north); //Offset startindex
      copy_to_from_ghostBuffer.setArg(22,1); //Direction

      NDRange global(ghost_element_send_north);

      queue.enqueueNDRangeKernel(copy_to_from_ghostBuffer,NullRange,global,NullRange,NULL,&copy_to_host_events[0]);
      readBuffer(queue,d_north_send_buffer,north_send_buffer,ghost_element_send_north * N_VEC * sizeof(real_t),&copy_to_host_events[1]);
    }
    //Coppying south layer
    if (ghost_layers_south != 0){

      //Write the host data to the buffer

      copy_to_from_ghostBuffer.setArg(19,d_south_send_buffer); //Buffer
      copy_to_from_ghostBuffer.setArg(21,(total_voxels - ghost_element_recive_south - ghost_element_send_south )); //Offset startindex
      copy_to_from_ghostBuffer.setArg(22,1); //Direction

      NDRange global(ghost_element_send_south);

      queue.enqueueNDRangeKernel(copy_to_from_ghostBuffer,NullRange,global,NullRange,NULL,&copy_to_host_events[2]);
      readBuffer(queue, d_south_send_buffer,south_send_buffer,ghost_element_send_south * N_VEC * sizeof(real_t),&copy_to_host_events[3]);
    }

  } catch(Error error) {
    std::cout << "LBM " << index << " copy to cpu" << std::endl;
    std::cout << error.what() << "(" << error.err() << ")" << std::endl;
    std::cout << getCLErrorString(error.err()) << std::endl;
    PAUSE
    exit(-1);
  }

}

real_t* LBM_3D::getNorthSendBuffer(){
  return north_send_buffer;
}
real_t* LBM_3D::getSouthSendBuffer(){
  return south_send_buffer;
}

void LBM_3D::setNorthNeighbour(LBM_3D *north_neighbour, int north_rank)
{
  this->north_neighbour = north_neighbour;
  this->north_rank = north_rank;

#if MODE == mode_mpi || MODE == mode_hybrid
  if (ghost_layers_north != 0 && north_rank != -1){
    MPI_Send_init(north_send_buffer,ghost_element_send_north * N_VEC,MPI_FLOAT,north_rank,0,MPI_COMM_WORLD,&send_north);
    MPI_Recv_init(north_recv_buffer,ghost_element_recive_north* N_VEC,MPI_FLOAT,north_rank,0,MPI_COMM_WORLD,&reciv_north);
  }
#endif
}

void LBM_3D::setSouthNeighbour(LBM_3D *south_neighbour, int south_rank)
{
  this->south_neighbour = south_neighbour;
  this->south_rank = south_rank;

#if MODE == mode_mpi || MODE == mode_hybrid 
  if(ghost_layers_south != 0 && south_rank != -1){
    MPI_Send_init(south_send_buffer,ghost_element_send_south* N_VEC,MPI_FLOAT,south_rank,0,MPI_COMM_WORLD,&send_south);
    MPI_Recv_init(south_recv_buffer,ghost_element_recive_south* N_VEC,MPI_FLOAT,south_rank,0,MPI_COMM_WORLD,&reciv_south);
  }
#endif
}

void LBM_3D::finish(int place){
	
	try{
		queue.finish();
	} catch(Error error) {
		std::cout << "finish" << std::endl;
		std::cout << error.what() << "(" << error.err() << ")" << std::endl;
		std::cout << getCLErrorString(error.err()) << std::endl;
		exit(-1);
	}



	if ( place & COLLIDE){
		double tmp = 0.0;
		
		tmp = executionTime(collide_events[0]);
		time_collide_inner += tmp;
		time_collide_last_itter = tmp;

		if (ghost_layers_north != 0){
			tmp = executionTime(collide_events[1]);
			time_collide_ghost += tmp;
			time_collide_last_itter += tmp;
		}
		if (ghost_layers_south != 0){
			tmp = executionTime(collide_events[2]);
			time_collide_ghost += tmp;
			time_collide_last_itter += tmp;
		}
		
		time_collide += time_collide_last_itter;
	}

	if ( place & STREAM){
		time_stream_last_itter = executionTime(stream_event);
		time_stream += time_stream_last_itter;
	}

	if ( place & AVG_VEL){
		time_avg_velocity += executionTime(avrage_events[0]);
		time_avg_velocity += executionTime(avrage_events[1]);
	}

	if ( place & COPY_TO_CPU){
		if (ghost_layers_north != 0){
			time_copy_to_cpu += executionTime(copy_to_host_events[0]);
			//time_copy_to_cpu_1 += executionTime(copy_to_host_events[0]);

			time_copy_to_cpu += executionTime(copy_to_host_events[1]);
			//time_copy_to_cpu_2 += executionTime(copy_to_host_events[1]);
		}
		if (ghost_layers_south != 0){
			time_copy_to_cpu += executionTime(copy_to_host_events[2]);
			//time_copy_to_cpu_1 += executionTime(copy_to_host_events[2]);
			time_copy_to_cpu += executionTime(copy_to_host_events[3]);
			//time_copy_to_cpu_2 += executionTime(copy_to_host_events[3]);
		}
	}

  if ( place & COPY_TO_GPU){
    if (ghost_layers_north != 0){
      time_copy_to_gpu += executionTime(copy_to_device_events[0]);
      time_copy_to_gpu += executionTime(copy_to_device_events[1]);
    }
    if (ghost_layers_south != 0){
      time_copy_to_gpu += executionTime(copy_to_device_events[2]);
      time_copy_to_gpu += executionTime(copy_to_device_events[3]);
    }
  }

  total_lbm_time = time_collide + time_stream + time_avg_velocity + time_copy_to_cpu + time_copy_to_gpu + time_exchange;

}



int LBM_3D::getActiveElements()
{
   return total_voxels - ghost_element_recive_north - ghost_element_recive_south;
}


int LBM_3D::copyAvgVelToCPU(real_t *u_new)
{
  Event e;
  int antElemnts = getActiveElements();
  readBuffer(queue,
    d_avrage_velocity, 
    ghost_element_recive_north * sizeof(real_t), 
    u_new,
    antElemnts*sizeof(real_t),
    &e);
  queue.finish();
  return antElemnts;
}


void LBM_3D::validation()
{
  writeVelocityFieldToFile();
}

void LBM_3D::writeVelocityFieldToFile()
{
  int local = 41;
  //int local = 28;
  //int local = 21;

  FILE *pFile;
  std::string file_n;
  file_n += LOGFOLDER;
  file_n += "vel";


  stringstream out_index;
  out_index << index;
  file_n += "_" + out_index.str();


  file_n += ".txt";
  pFile = fopen(file_n.c_str(),"a+");

  for(int z=0; z<getSizeZ(); z++){
    fprintf(pFile, "layer %d\n", (index * local ) + z);
    printf("layer %d\n", (index * local ) + z);
    for(int y=0; y<getSizeY(); y++){
      for(int x=0; x<getSizeX(); x++){
        fprintf(pFile, "%e ", u_new[getIndex(x,y,z)]);	
        printf("%e ", u_new[getIndex(x,y,z)]);	
      }
      fprintf(pFile, "\n");
      printf("\n");
    }
    fprintf(pFile, "\n");
    printf("\n");
  }

  fclose(pFile);

  /*
  FILE *pFile;
  pFile = fopen("logs\\profile.txt","a+");
  fprintf(pFile, "\n\n : (%d,%d,%d) \n", getSizeX(), getSizeY(), getSizeZ());
  fprintf(pFile, "Iteration: %d \n", iter_count);

  for(int x=0; x<getSizeX(); x++)
  {
  fprintf(pFile, "x: %x \n", x);

  for(int y=0; y<getSizeY(); y++)
  {
  for(int z=0; z<getSizeZ(); z++)
  {
  fprintf(pFile, "%f ", u_new[getIndex(x,y,z)]);	
  }
  fprintf(pFile, "\n");
  }

  fprintf(pFile, "\n");
  }

  fprintf(pFile, "\n");
  fclose(pFile);*/

}

void LBM_3D::writeRight1DProfileToFile()
{
  FILE *pFile;
  pFile = fopen("logs\\right_1d_profile.txt","a+");
  fprintf(pFile,"Iteration: %d \n", iter_count);
  fprintf(pFile,"Viscosity: %f \n", viscosity);
  fprintf(pFile,"Max Velocity: %f \n", u_max);

  real_t r = (( (real_t)getSizeY()) - 2 ) / 2.;
  real_t g = (force/(2*viscosity*19.));

  long double total_error = 0.0f;

  for(int iY=1; iY<getSizeY()-1; iY++)
  {		
    double correct = g * (pow(r, 2) - pow(((real_t) iY - 0.5 - r), 2));

    real_t our = u_new[getIndex(getSizeX()/2, iY, getSizeZ()/2)];
    long double real_error = (correct  - our);
    total_error += pow(correct  - our, 2) / ((real_t) getSizeY()-2);

    fprintf(pFile, "%f %f %6.16Lf \n", our, correct, real_error);       
  }

  fprintf(pFile,"Error: %f\n", sqrtf(total_error));
  fprintf(pFile,"Error: %6.16f\n", sqrtf(total_error));
  fprintf(pFile,"Error: %e\n", sqrtf(total_error));
  fprintf(pFile,"\n");
  fclose(pFile);

  printf("Error: %f\n", sqrtf(total_error));
}

real_t LBM_3D::writeRight2DProfileToFile()
{	
  FILE *pFile;
  pFile = fopen("logs\\right_2d_profile.txt","a+");
  fprintf(pFile, "Iteration: %d \n", iter_count);
  fprintf(pFile, "Viscosity: %f \n", viscosity);
  fprintf(pFile,"Max Velocity: %f \n", u_max);

  real_t a = (real_t) (getSizeZ()-2)/2.;
  fprintf(pFile,"b: %f \n", a);

  real_t b = (real_t) (getSizeY()-2)/2.; 
  fprintf(pFile,"a: %f \n", b);

  real_t g = (force/(2*viscosity*19.));

  fprintf(pFile,"g: %f \n", g);

  real_t sum = 0.0f;
  real_t before = 0.0f;
  real_t er = 0.0f;
  real_t averageVelocity = 0.0f;
  real_t error = 0.0f;

  int numberOfElements = 0;

  for(int iY=0; iY<=b; iY++)
  {
    for(int iZ=0; iZ<=a; iZ++)
    {
      sum = 0.0f;
      before = 0.0f;

      for(int n=0; n<10; n++)
      {
        real_t n1 = (real_t)(2*n+1);
        real_t b2 = 2.f*b;

        real_t a_1 = pow(-1., n+1);
        real_t a_2 = cosh((n1 * 3.14f * (real_t) iZ) / b2);

        real_t a_4 = cos((n1 * 3.14f * (real_t) iY) / b2);

        real_t b_1 = pow(n1, 3);
        real_t b_2 = cosh((n1 * 3.14f * a) / b2);

        sum += ( a_1 * a_2 * a_4  ) / ( b_1 * b_2  );
      }

      before = pow(b, 2) - pow((real_t)iY, 2); 	
      real_t real_v =  g * ((before + (32.f * pow(b, 2) / pow(3.14f, 3)) * sum ));	

      averageVelocity += real_v;

      real_t our = u_new[getIndex(getSizeX()/2, (iY+(getSizeY()/2)), (iZ+(getSizeZ()/2)))];

      er += pow(real_v  - our, 2) / (a*b);
      error += (real_v  - our);

      fprintf(pFile, "(%lf,%lf) ", our, real_v);	

      numberOfElements++;
    }
    fprintf(pFile, "\n");
  }	

  fprintf(pFile,"\nNumber Of Elements: %d \n", numberOfElements);
  fprintf(pFile,"Error: %lf\n", sqrtf(er));
  fprintf(pFile,"Real error: %lf\n", error);
  fprintf(pFile,"\n");
  fclose(pFile);

  return sqrtf(er);
}


void LBM_3D::printState()
{
  if(!(iter_count % iter_save) && iter_count > 0)
  {	
    printf("%d (%ds) ", iter_count, (int) total_lbm_time);
    printf("Convergence: %e ", convergence);
    //printf("MLUPS: %f (%f) ", MLUPS, best_mlups);
    printf("AvgVel: %e ", avg_velocity);
    //		printf("Perm: %lf ", permeability_new);
    printf("LBM: %fls ", time_lbm);
    printf("Collide: %fs ", time_collide);
    printf("Stream: %fs ", time_stream);	
    printf("Lines: %fs ", time_lines);
    printf("Avg Vel: %fs \n", time_avg_velocity);
  }

}

void LBM_3D::printInitalState()
{
  // Print simulation to console
  printf("\n\nLBM: %d\n", index);
  printf("Dimension: (%d,%d,%d) \n", getSizeX(), getSizeY(), getSizeZ());

  printf("Ghost North:\n");
  printf("\tLayers: %d \n", ghost_layers_north);
  //printf("\tMPI rank: %d \n", north_rank);
  printf("\tSend elements: %d \n", ghost_element_send_north);
  printf("\tRecive elements: %d \n", ghost_element_recive_north);

  printf("Ghost South:\n");
  printf("\tLayers: %d \n", ghost_layers_south);
  //printf("\tMPI rank: %d \n", south_rank);
  printf("\tSend elements: %d \n",ghost_element_send_south);
  printf("\tRecive elements: %d \n",ghost_element_recive_south);


  printf("Active lattice locations (own): %d \n", (int) voxelsActive);
  printf("Active lattice locations (including ghost and extra): %d \n", total_voxels);

  printf("Omega: %lf \n", omega);
  printf("Kinematic Viscosity: %f\n", viscosity);
  //printf("Porosity: %f\n", porosity);
  printf("Force: %f \n", force);
  printf("Memory Usage:\n");
  printf("\tGPU: %f MB\n",(memoryUsageGPU /pow(1024.,2)));
  printf("\tCPU: %f MB\n",(memoryUsageCPU /pow(1024.,2)));

  if(sizeof(real_t)==4)
    printf("32-bit Precision \n");
  else
    printf("64-bit Precision \n");

  // Write simulation to file
  fprintf(pFile, "\n\nLBM: %d\n", index);
  fprintf(pFile, "Dimension: (%d,%d,%d) \n", getSizeX(), getSizeY(), getSizeZ());

  fprintf(pFile, "Ghost North:\n");
  fprintf(pFile, "\tLayers: %d \n", ghost_layers_north);
  fprintf(pFile, "\tSend elements: %d \n", ghost_element_send_north);
  fprintf(pFile, "\tRecive elements: %d \n", ghost_element_recive_north);

  fprintf(pFile, "Ghost South:\n");
  fprintf(pFile, "\tLayers: %d \n", ghost_layers_south);
  fprintf(pFile, "\tSend elements: %d \n",ghost_element_send_south);
  fprintf(pFile, "\tRecive elements: %d \n",ghost_element_recive_south);


  fprintf(pFile, "Active lattice locations (own): %d \n", (int) voxelsActive);
  fprintf(pFile, "Active lattice locations (including ghostand extra): %d \n", total_voxels);

  fprintf(pFile, "Omega: %lf \n", omega);
  fprintf(pFile, "Kinematic Viscosity: %f\n", viscosity);
  //fprintf(pFile, "Porosity: %f\n", porosity);
  fprintf(pFile, "Force: %f \n", force);
  fprintf(pFile, "Memory Usage:\n");
  fprintf(pFile, "\tGPU: %f MB\n",(memoryUsageGPU /pow(1024.,2)));
  fprintf(pFile, "\tCPU: %f MB\n",(memoryUsageCPU /pow(1024.,2)));

  if(sizeof(real_t)==4)
    fprintf(pFile, "32-bit Precision \n");
  else
    fprintf(pFile, "64-bit Precision \n");

}

void LBM_3D::printConvergedState()
{	

  validation();

  // Write simulation to file
  fprintf(pFile, "Simulation Finished After: %d Iterations And %f Seconds \n", iter_count, total_lbm_time);
  fprintf(pFile, "Average MLUPS: %f\n", MLN / (total_lbm_time/iter_count));
  fprintf(pFile, "Maximum MLUPS: %f\n", best_mlups);

  FILE *p_mlups;
  p_mlups = fopen("logs\\MLUPS.txt","a+");
  fprintf(p_mlups, "%f \n", total_mlups/iter_count);
  fclose(p_mlups);

  FILE *p_time;
  p_time = fopen("logs\\TIME.txt","a+");
  fprintf(p_time, "%f \n", total_lbm_time);
  fclose(p_time);
}


int LBM_3D::printTimings(char *stream,int initOffsett){
	int offset = initOffsett;
	offset += sprintf(stream+offset, "LBM %d:\n",index);

	offset += sprintf(stream+offset, "\tSetup:\n");
	offset += sprintf(stream+offset, "\t\tAllocateMemory\t %f \n",time_allocate_memory);
	offset += sprintf(stream+offset, "\tSetup total\t\t %f\n",time_allocate_memory);

	offset += sprintf(stream+offset, "\n\tComputation:\n");
	//offset += sprintf(stream+offset, "\t\tCollide\t\t %f \n",time_collide);
	offset += sprintf(stream+offset, "\t\tCollide_inner\t %f \n",time_collide_inner);
	offset += sprintf(stream+offset, "\t\tCollide_ghost\t %f \n",time_collide_ghost);
	offset += sprintf(stream+offset, "\t\tStream\t\t %f \n",time_stream);
	offset += sprintf(stream+offset, "\t\tConvrgence_copy\t %f \n",time_copy_to_cpu_convergence);
	offset += sprintf(stream+offset, "\t\tConvrgence\t %f \n",time_convergence_check);



	offset += sprintf(stream+offset, "\t\tAvgVelocity\t %f \n",time_avg_velocity);
	offset += sprintf(stream+offset, "\tComputation total\t %f\n",time_collide+time_stream+time_avg_velocity + time_copy_to_cpu_convergence + time_convergence_check);


	offset += sprintf(stream+offset, "\n\tComunication:\n");
	offset += sprintf(stream+offset,"\t\tCopyToCPU\t %f \n",time_copy_to_cpu);
	//fprintf(stream, "\t\tCopyToCPU_1\t %f \n",time_copy_to_cpu_1);
	//fprintf(stream, "\t\tCopyToCPU_2\t %f \n",time_copy_to_cpu_2);
	offset += sprintf(stream+offset, "\t\tCopyToGPU\t %f \n",time_copy_to_gpu);
	offset += sprintf(stream+offset, "\t\tExchange\t %f \n",time_exchange);
	//time_copy_to_cpu = time_copy_to_cpu_1 + time_copy_to_cpu_2;
	offset += sprintf(stream+offset, "\tComunication total\t %f\n",time_copy_to_cpu+time_copy_to_gpu+time_exchange);

	offset += sprintf(stream+offset, "\n\tLBM TOTAL \t\t %f \n",total_lbm_time);
	offset += sprintf(stream+offset, "\tLBM TOTAL WALL \t\t %f \n",total_lbm_wall_time);

	return offset;

}

void LBM_3D::printMLUPS(FILE *stream){

  //fprintf(stream, "LBM %d: MLUPS: %f (%f / %e) \n",index, (voxelsActive / 1.0e6) / (time_collide_last_itter + time_stream_last_itter), voxelsActive / 1.0e6, (time_collide_last_itter + time_stream_last_itter));
  fprintf(stream, "LBM %d: MLUPS: %f \n",index, (voxelsActive / 1.0e6) / (time_collide_last_itter + time_stream_last_itter));

}

void LBM_3D::createFileName()
{
  std::string file_n;
  file_n += "GPU";

  if(sizeof(real_t) == 4)
    file_n += "_32";
  else
    file_n += "_64";

  stringstream out_x;
  out_x << getSizeX();
  file_n += "_" + out_x.str();

  stringstream out_y;
  out_y << getSizeY();
  file_n += "_" + out_y.str();

  stringstream out_z;
  out_z << getSizeZ();
  file_n += "_" + out_z.str();

  stringstream out_index;
  out_index << index;
  file_n += "_" + out_index.str();


  std::string dummy;
  file_n_time = LOGFOLDER + file_n + "_time.txt";
  //file_n_convergence = LOGFOLDER + file_n + "_conv.txt";
  file_n_convergence = LOGFOLDER + dummy + "conv.txt";
  file_n_permeability = LOGFOLDER + file_n + "_perm.txt";
  file_n_MLUPS = LOGFOLDER + file_n + "_MLUPS.txt";
  file_n_vel = LOGFOLDER + file_n + "_Vel.txt";
}
