#ifndef LBM_3D_CPU_H
#define LBM_3D_CPU_H

//#include "Shader.h"
#include <ctime>
//#include <windows.h>


#include <sstream>

#include <omp.h>

//include mpi.h
#include <mpi.h>

// C library to perform input and output operations
#include <stdio.h>

// C standard general utilities library 
#include <stdlib.h>

// C numerics library
#include <math.h>	

// The lattice Boltzmann method
#include "Config.h"
#include "Types.h"
//#include "MarchingCubes.h"

// OpenCl imp of Lattice boltzman
#include "OpenCl_utils.h"


#include <iostream>
#include <string>

using namespace std;

class LBM_3D
{
public:
	LBM_3D(int Id, uchar *data, real_t omegaInit, int sizeX, int sizeY, int sizeZ, int ext, 
    int ghostLayersNorth, int ghostLayersSouth, bool writeToFile, CommandQueue inputQueue, real_t forceInit);

	~LBM_3D(); 

////////////////////////////////////////////////////////////////////////////////
//! Public functions
////////////////////////////////////////////////////////////////////////////////
	// Get physical dimension in x 
	int getSizeX();

	// Get physical dimension in y
	int getSizeY();

	// Get physical dimension in z
	int getSizeZ();

	// Execute one single LBM iteration
	//void doIteration(real_t *local_avg_vel);

  void doCollide();

  void doCollide_inner();
  void doCollide_ghost();

  void doStream();

  void calculateAvrageVelocity(real_t *local_avg_vel);

  void getMaxParticleMovment(real_t *max_movment);


	
	void printInitalState();

	void printState();

	void printConvergedState();

	//void printTimings(FILE *stream);
	int printTimings(char *stream,int initOffsett);
	void printMLUPS(FILE *stream);

	void createFileName();

	// Write the velocity field to file
	void writeVelocityFieldToFile();


  //void exhangeGhostLayers_threaded(LBM_3D *north_neighbour, LBM_3D *south_neighbour);
  //void exhangeGhostLayers_MPI(int north_rank, int south_rank);

  void exhangeGhostLayers_threaded();
  void exhangeGhostLayers_MPI();

  void copyGhostLayersToGPU();
  void copyGhostLayersToCPU();

  void finish(int place);

  real_t* getNorthSendBuffer();
  real_t* getSouthSendBuffer();

  int copyAvgVelToCPU(real_t *u_new);
  int getActiveElements();

  void setNorthNeighbour(LBM_3D *north_neighbour, int north_rank); 
  void setSouthNeighbour(LBM_3D *south_neighbour,int south_rank);

  double total_lbm_wall_time;

private:

////////////////////////////////////////////////////////////////////////////////
//! Private functions
////////////////////////////////////////////////////////////////////////////////

	// Get solid value at location x, y, and z
	int getSolid(int x, int y, int z);


	// Measure of empty space in a material
	//real_t getPorosity();


	// Get index at location x, y, and z
	int getIndex(int x, int y, int z);

	
	// Calculate velocity profile error
	void calculateVelocityProfileError();


	//Write the right velocity field to file
	void writeRight1DProfileToFile();
	real_t writeRight2DProfileToFile();
	

	//void calculateActiveVoxels();

	void validation();

	void computeTotalVoxels();

  void createHostDataStructureAndCalculateTotalAndActiveVoxels(uchar *data, int ghostLayersNorth, int ghostLayersSouth);
  void calculateGhostElemetsNorthAndSouth(int ghostLayersNorth, int ghostLayersSouth);

  //double executionTime(Event &event);
		
////////////////////////////////////////////////////////////////////////////////
//! Variables
////////////////////////////////////////////////////////////////////////////////

  int ghost_layers_north;
  int ghost_element_recive_north;
  int ghost_element_send_north;
  real_t *north_send_buffer;
  real_t *north_recv_buffer;

  int ghost_layers_south;
  int ghost_element_recive_south;
  int ghost_element_send_south;
  real_t *south_send_buffer;
  real_t *south_recv_buffer;

  // Used to save the north and south LBM for both MPi and threaded version
  LBM_3D *north_neighbour;
  LBM_3D *south_neighbour;
  int north_rank;
  int south_rank;


  //Used for handled in MPI

  MPI_Request send_north;
  MPI_Request send_south;

  MPI_Request reciv_north;
  MPI_Request reciv_south;


  
  // Number of active lattice locations in the domain
  double voxelsActive;

  // Number of active lattices in the domain (including ghost layer and extra)
  int total_voxels;


  int index;

  unsigned long memoryUsageGPU;
  unsigned long memoryUsageCPU;

  // Force to add to each lattice location
  real_t force;

	// Physical dimension in x
	int size_x;

	// Physical dimension in y
	int size_y;

	// Physical dimension in z
	int size_z;


	// Relaxation parameter
	real_t omega;

	// Iteration counter
	int iter_count; 
	
	
	// Number of iterations between printouts
	int iter_save; 

	// Host memory for velocity fields
	real_t *u_new;

  //Host memory for the max Movment for particles
  real_t *max_movments;

	// Maximum velocity
	real_t u_max;
	real_t u_max_old;

	// Host memory for solids 
	int *solid_h; 

  // Porosity is a measure of the void spaces in a material
  //double porosity;

  // Viscosity is the quantity that describes a fluid's resistance to flow
  double viscosity;



	// Millions lattice locations
	double MLN;

	// Extra space before and after the porous rocks
	int extra;


	// Simulation log file
	FILE *pFile;
	
	// The total simulation time
	double total_lbm_time;


	
	// Write to file on/off
	bool write_to_file;

	// The privious iterations average velocity
	double last_avg_velocity;
	double avg_velocity;
	double convergence;
	double MLUPS;


	// The best MLUPS so farfloat_t
	double best_mlups;

	// Used for timeing of the total simulation time
	double timer;

	double time_collide;

	double time_collide_inner;
	double time_collide_ghost;

	double time_stream;
	double time_collide_last_itter;
	double time_stream_last_itter;

	double time_lbm;
	double time_lines;
	double time_avg_velocity;
	double time_allocate_memory;
	double time_copy_to_cpu;
	double time_copy_to_gpu;
	double time_exchange;

  double time_convergence_check;
  double time_copy_to_cpu_convergence;


	//int hits;	

	std::string file_n_time;
	std::string file_n_convergence;
	std::string file_n_permeability;
	std::string file_n_MLUPS;
	std::string file_n_vel;

	double total_mlups;

  CommandQueue queue;
  Context context;

  Buffer d_f[N_VEC];
  Buffer d_avrage_velocity;

  Buffer d_north_send_buffer;
  Buffer d_north_recv_buffer;

  Buffer d_south_send_buffer;
  Buffer d_south_recv_buffer;


  Image3D d_solids;

  Kernel colide_and_swap;
  Kernel stream_and_swap;
  Kernel avrage_velocity;
  Kernel reduction;
  Kernel copy_to_from_ghostBuffer;


	/*
	 * 0 inner
	 * 1 ghost north
	 * 2 ghost south
	 */
  vector<Event> collide_events;
  
  Event stream_event;

  /*
   * 0 Avrage velocity kernel
   * 1 Reduce kernel
   */

  vector<Event> avrage_events;


  /*
   * Using 4 of these events 
   * 0 kernel north
   * 1 buffer north
   * 2 kernel south
   * 3 buffer south
   */
  vector<Event> copy_to_device_events; 
  
  /*
   * Using 4 of these events 
   * 0 buffer north
   * 1 kernel north
   * 2 buffer south
   * 3 kernel south
   */
  vector<Event> copy_to_host_events;
  
};

enum finish_timing {
  COLLIDE     =  1,
  STREAM      =  2,
  AVG_VEL     =  4,
  COPY_TO_CPU =  8,
  COPY_TO_GPU = 16
};

#endif
