#ifndef OPENCL_UTIL_H_
#define OPENCL_UTIL_H_

#define __NO_STD_VECTOR // Use cl::vector instead of STL version
#define __CL_ENABLE_EXCEPTIONS

#include <iostream>
#include <fstream>
#include <utility>

// The lattice Boltzmann method
#include "Config.h"
#include "Types.h"
#include "Timer.h"

#include "cl.hpp"


using namespace cl;

void setupOpenCL(CommandQueue *queues, int ant_device, int startpos);

char * getCLErrorString(cl_int error);

void readBuffer(CommandQueue queue, Buffer src, void *dst, int length_in_byte,Event *event);
void readBuffer(CommandQueue queue, Buffer src,  int offset,void *dst, int length_in_byte,Event *event);

void writeBuffer(CommandQueue queue, Buffer dst, void *src, int length_in_byte,Event *event);
void writeBuffer(CommandQueue queue, Buffer dst,int offset, void *src, int length_in_byte,Event *event);

Buffer createBuffer(int size, cl_mem_flags flags, void* data);
Buffer createBuffer(int size, cl_mem_flags flags);

Kernel createKernel(const char * filename);




void streamAndSwapGPU(CommandQueue queue, Kernel stream_and_swap, int size_x, int size_y, int size_z, Event *event, Buffer *d_f, Image3D &d_solids, Buffer &max_change );
void collideAndSwapGPU(CommandQueue queue, Kernel colide_and_swap, 
					   int size_x, int size_y, int size_z, 
					   int offset_x, int offset_y, int offsett_z,
					   real_t omega, real_t force, Event *event, Buffer *d_f, Image3D &d_solids, int index);

void allocateMemoryGPU(int size_x, int size_y, int size_z, int totalVoxels, int *solid_h, Buffer *d_f, Buffer &d_avrage_velocity, Image3D &d_solids, double &time);
void freeMemoryGPU();



void getAverageVelocityXGPU(CommandQueue queue, Kernel avrage_velocity, Kernel reduction,  Buffer *d_f, Buffer &d_avrage_velocity, Image3D &d_solids,
                            int total_voxels, int size_x, int size_y, int size_z, int extra, vector<Event> &evens, int ghost_layers_north, int ghost_layers_south, real_t *avg_vel);

double executionTime(Event &event);

#endif