
#include "Simulation_mpi.h"


extern double avg_velocity;
extern double last_avg_velocity;
extern double permeability_old;
extern double permeability_new;

extern uchar *data;
extern double convergence;
extern double viscosity;
extern real_t porosity;
extern int iter_count;

extern double MLUPS;
extern double MLN;
extern long long int ActivevoxelCount;

//Timings 
extern double time_total;
extern double time_lbm;
extern double time_setup;
extern double time_wait_to_finish;
extern double time_wait_to_finish_comm;
extern double time_reduce;
extern double time_iteration;

extern int size_x;
extern int size_y;
extern int size_z;
extern int extra;

extern real_t prev_max_movment;
extern real_t max_movment;
extern double rate;


  




extern int *solid_size;
extern int *solid_offset;

//used in both run simulation and print timings
extern LBM_3D **lbm;


/*
static void
test_mask ( cpu_set_t *m, int nprocs, pid_t mypid )
{
  {
    typedef unsigned long size_t;
    //sched_getaffinity ( mypid, nprocs, m );
    printf ( "pid %d reports %d processors on this computer\n", mypid, nprocs );
    for ( int cpu=0; cpu<nprocs; cpu++ )
      printf ( "ISSET(%d,mask) = %d\n", cpu, CPU_ISSET(cpu,m) );
  }
}
*/


void runSimulationMPI(int mpi_rank, int mpi_size, int num_threads)
{
#if MODE == mode_mpi || MODE == mode_hybrid  

  {

  typedef unsigned long size_t;
  pid_t mypid;
  int nprocs;
  nprocs = sysconf ( _SC_NPROCESSORS_CONF );  // Linux specific(?)
  cpu_set_t mask;
  mypid = getpid();

  //clear the mask for each core
  for ( int i=0; i<nprocs; i++ )
    CPU_CLR ( i, &mask );
  //test_mask( &mask, nprocs, mypid );

  //CPU_SET ( mpi_rank*4, mask );
  CPU_SET ( mpi_rank, &mask );
  //CPU_SET ( 0, &mask );
  sched_setaffinity ( mypid, nprocs, &mask );

  //read the cpu back to verify
  //sched_getaffinity ( mypid, nprocs, &mask );
  //test_mask( &mask, nprocs, mypid );
  
  printf("Hello from node %d of %d on pid %d with cpu mask ",mpi_rank+1,mpi_size,mypid);
  for ( int cpu=0; cpu<nprocs; cpu++ )
      printf ( "%d ",CPU_ISSET(cpu,&mask) );
  printf("\n");

  }

  uchar *node_data_ptr;
  uchar **local_data_ptr;
  CommandQueue *queues;


  //int mpi_rank, mpi_size;
  //int num_threads = 1;
  int mpi_north_rank;
  int mpi_south_rank;


  //timings
  long t1;
  long total_start;
  long t2;
  long iteration_start;
  long iteration_end;

  int converged = 0;



  //used to distrubute the stone size
  int *stone_size = new int[4];

  int *all_ghost_layers_north = new int [mpi_size];
  int *all_ghost_layers_south = new int [mpi_size];

  //int local_size_x;
  //int local_size_y;
  int node_size_z;
  int node_extra;
  int *node_sizes_z;

  int local_size_z;

  //real_t local_avg_velocity;
  real_t *local_avg_velocities;
  real_t omega;
  real_t prev_max = 0.0;
  real_t *local_max_movment;
  real_t local_max;


  int ghost_layers_north = 0;
  int ghost_layers_south = 0;


  int hits = 0;


    getTimeInMs(t1);
    getTimeInMs(total_start);

  
  //Fill the send buffer on rank 0
  if(mpi_rank == 0){
    stone_size[0] = getSizeX();
    stone_size[1] = getSizeY();
    stone_size[2] = getSizeZ();
    stone_size[3] = extra;
  }

  //Distrubute the stone size, collect all ints so we only do one bcast call
  MPI_Bcast(stone_size,4,MPI_INT,0,MPI_COMM_WORLD);


  //Set the stone size on all nodes
  size_x = stone_size[0];
  size_y = stone_size[1];
  size_z = stone_size[2];
  node_extra = stone_size[3];


  //Distrubute the stone amongst the nodes via scatterv
  lbm = new LBM_3D *[num_threads];
  local_avg_velocities = new real_t[num_threads]; 
  local_max_movment = new real_t[num_threads];
  local_data_ptr =  (uchar**) malloc(num_threads * sizeof(uchar*));

  solid_size = new int[mpi_size];
  solid_offset = new int [mpi_size];
  node_size_z =  ceil( (double)getSizeZ() / (double) mpi_size);
  node_sizes_z =  new int[mpi_size];

  omega = 1.f / tau;
  viscosity = (1./6. * (2./omega - 1.)); 



  mpi_north_rank = (mpi_rank + mpi_size - 1) % mpi_size;
  mpi_south_rank = (mpi_rank + 1) % mpi_size;


  
  
  int startPos = 0;

#if MODE == mode_hybrid
  startPos = 0;
  //startPos = mpi_rank;
#elif MODE == mode_mpi
  startPos = mpi_rank % N_MPI_PRO_PER_NODE;
#endif


  queues = new CommandQueue[num_threads];
  //setupOpenCL(queues,num_threads,mpi_rank);
  //setupOpenCL(queues,num_threads,0);
  setupOpenCL(queues,num_threads,startPos);


#if PERIODIC_Z
  solid_offset[0] = GHOST_LAYERS;
#else
 solid_offset[0] = 0;
#endif

 
  for(int i = 0; i < mpi_size; i++){
  
    if (!PERIODIC_Z && i == 0)
      all_ghost_layers_north[i] = 0;
    else
      all_ghost_layers_north[i] = GHOST_LAYERS;  
    
    if (!PERIODIC_Z && i == mpi_size -1)
      all_ghost_layers_south[i] = 0;
    else
      all_ghost_layers_south[i] = GHOST_LAYERS;




    solid_size[i] = (i != (mpi_size - 1)) ? node_size_z : (getSizeZ() - (mpi_size - 1) * node_size_z);
    node_sizes_z[i] = (i != (mpi_size - 1)) ? node_size_z : (getSizeZ() - (mpi_size - 1) * node_size_z);
    if (i != mpi_size -1){
      solid_offset[i + 1] = solid_offset[i] + solid_size[i];
    }

    //recalculate the size and offsett to be in ant element insted of ant rows in z as scatterv needs this
    solid_size[i] = (solid_size[i] + all_ghost_layers_north[i] + all_ghost_layers_south[i]) * getSizeX() * getSizeY();
    solid_offset[i] = (solid_offset[i] - all_ghost_layers_north[i]) * getSizeX() * getSizeY();

  }

  //Each node find its own ghost layers
  ghost_layers_north = all_ghost_layers_north[mpi_rank];
  ghost_layers_south = all_ghost_layers_south[mpi_rank];

  
  //printf("size: %d off: %d gn: %d gs %d\n",solid_size[mpi_rank],solid_offset[mpi_rank],ghost_layers_north,ghost_layers_south);

  node_data_ptr = (uchar *) malloc(solid_size[mpi_rank] * sizeof(uchar));
  MPI_Scatterv (data,solid_size,solid_offset,MPI_CHAR,node_data_ptr,solid_size[mpi_rank],MPI_CHAR,0,MPI_COMM_WORLD);




  //Distribute the data amongst the GPUs on the node
  local_size_z = ceil((double) node_sizes_z[mpi_rank] / (double) num_threads);
  int ghost_layer_tmp = ghost_layers_north;
  uchar *data_ptr;
  
  if (ghost_layer_tmp != 0){
    data_ptr = &node_data_ptr[ghost_layer_tmp * getSizeX() * getSizeY()];
  }else{
    data_ptr = node_data_ptr;
  }

  int *local_solid_offset = new int[num_threads];
  int *local_solid_size = new int[num_threads];

  local_solid_offset[0] = 0;
  for(int i = 0; i < num_threads; i++){
    local_solid_size[i] = (i != (num_threads - 1)) ? local_size_z : (node_sizes_z[mpi_rank] - local_solid_offset[i]);

    //Static work loading when using the hybrid working model
/*
#if MODE == mode_hybrid
    if (i==0){
      local_solid_size[i] *= 0.90; 
    }
#endif
*/    
    local_data_ptr[i] = &data_ptr[(local_solid_offset[i] - ghost_layer_tmp) * getSizeX() * getSizeY()];

    if (i != num_threads -1){
      local_solid_offset[i + 1] = local_solid_offset[i] + local_solid_size[i];
    }

    ghost_layer_tmp = GHOST_LAYERS;
  }




  //_lbm = new LBM_3D(mpi_rank,local_data_ptr,omega,getSizeX(),getSizeY(),local_sizes_z[mpi_rank],
  //  local_extra,ghost_layers_north,ghost_layers_south,FALSE,queues[0],force);
  
//Init the LBM solvers
  for(int i = 0; i < num_threads; i++){

    int local_ghost_layers_north = 0;
    int local_ghost_layers_south = 0;


    if (!PERIODIC_Z && i == 0){
      local_ghost_layers_north = ghost_layers_north;
    }else{
      local_ghost_layers_north = GHOST_LAYERS;
    }

    if (! PERIODIC_Z && i == num_threads -1){
      local_ghost_layers_south = ghost_layers_south;
    }else{
      local_ghost_layers_south = GHOST_LAYERS;
    }

    //todo: tenk over om dette blir rett
    lbm[i] = new LBM_3D((mpi_rank * num_threads)+i ,local_data_ptr[i],omega,getSizeX(),getSizeY(),local_solid_size[i],
      node_extra,local_ghost_layers_north,local_ghost_layers_south,FALSE,queues[i],force);

    //lbm[i] = new LBM_3D(i,local_data_ptr[i], omega, getSizeX(), getSizeY(), solid_size[i], extra,
    //  ghost_layers_north, ghost_layers_south, write_to_file, queues[i],force);
  }

  //Have to set the north and south border after creation all all 
  for(int i = 0; i < num_threads; i++){

    LBM_3D *north_neighbour;
    LBM_3D *south_neighbour;
    if (!PERIODIC_Z && i == 0){
    north_neighbour = NULL;
    }else{
    north_neighbour = lbm[(i + num_threads -1) % num_threads];
    }

    if (!PERIODIC_Z && i == num_threads -1){
    south_neighbour = NULL;
    }else{
    south_neighbour = lbm[(i+1) % num_threads];
    } 

    lbm[i]->setNorthNeighbour(north_neighbour, -1);
    lbm[i]->setSouthNeighbour(south_neighbour, -1);

  }

   
  //set the Noth an south GPU as to have neighbour on the other node
  //this also is the only LBM when running in pure MPi mode so no internal swapping then
  lbm[0]->setNorthNeighbour(NULL, mpi_north_rank );
  lbm[num_threads -1]->setSouthNeighbour(NULL, mpi_south_rank );
  
  getTimeInMs(t2);
  time_setup += (double (t2 - t1) / toSec);

  while (iter_count < iterMax && !converged){

    MPI_Barrier(MPI_COMM_WORLD);
    //printf("Node %d starting itter %d\n",mpi_rank,iter_count);
    getTimeInMs(iteration_start);

    //Copy ghost layers to CPU
    for(int i = 0; i < num_threads; i++){
      lbm[i]->copyGhostLayersToCPU();
    }

    getTimeInMs(t1);
    waitToFinish(lbm,num_threads,COPY_TO_CPU);
    getTimeInMs(t2);
    time_wait_to_finish_comm += (double (t2 - t1) / toSec);

    for(int i = 0; i < num_threads; i++){
      lbm[i]->doCollide_inner();
    }

    //printf("Node %d dooing exchange\n",mpi_rank);
    for(int i = 0; i < num_threads; i++){
      //have to do both MPI and thread exchange to enable hybrid
      lbm[i]->exhangeGhostLayers_MPI();
      lbm[i]->exhangeGhostLayers_threaded();
    }   
    //printf("Node %d done exchange\n",mpi_rank);

    for(int i = 0; i < num_threads; i++){
      lbm[i]->copyGhostLayersToGPU();
    }
    
    getTimeInMs(t1);
    waitToFinish(lbm,num_threads,COPY_TO_GPU);
    getTimeInMs(t2);
    time_wait_to_finish_comm += (double (t2 - t1) / toSec);
    
    
    
    //MPI_Barrier(MPI_COMM_WORLD);
    //getTimeInMs(iteration_start);
   for(int i = 0; i < num_threads; i++){
    //lbm[i]->doCollide();
	  lbm[i]->doCollide_ghost();
    lbm[i]->doStream();
   }
    
    getTimeInMs(t1);
    waitToFinish(lbm,num_threads,COLLIDE|STREAM);
    getTimeInMs(t2);
    time_wait_to_finish += (double (t2 - t1) / toSec);
    
	MPI_Barrier(MPI_COMM_WORLD);
    getTimeInMs(iteration_end);
    
    
   time_iteration = (((double) (iteration_end - iteration_start)) / toSec);
    for(int i = 0; i < num_threads; i++){
      lbm[i]->total_lbm_wall_time += time_iteration;
    }

    if(!(iter_count % iterSave))
    {
      max_movment = 0.0;
      local_max = 0.0;
      for(int i = 0; i < num_threads; i++){
        lbm[i]->getMaxParticleMovment(&local_max_movment[i]);
        local_max = max(local_max, local_max_movment[i]);
      }
      MPI_Barrier(MPI_COMM_WORLD);
      getTimeInMs(t1);
      MPI_Reduce(&local_max,&max_movment,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
      getTimeInMs(t2);
      time_reduce += (double (t2 - t1) / toSec); 

    }
  

    if(mpi_rank == 0){
      //printf("Converged: %d %e \n", iter_count, convergence);
      //printf("%d Avg vel: %e\n",iter_count,avg_velocity);
      //avg_velocity = (avg_velocity / (double) ActivevoxelCount);
      //printf("node: %d %d local vel: %e\n",mpi_rank ,iter_count,local_avg_velocity);
      //printf("%d Avg vel: %e\n",iter_count,avg_velocity);

      /*permeability_new = getPermeability(avg_velocity, true);*/

      if(!(iter_count % iterSave))
      {

     
        rate =  (prev_max_movment / max_movment);
        convergence = fabs(1.0 - rate); //(prev_max_movment / max_movment));
        //printf("Converged: %d %e prev %e max %e rate %e  \n", iter_count, convergence,prev_max_movment, max_movment, rate);



        //convergence = getConvergence();
        //printf("Converged: %d %e \n", iter_count, convergence);


        if(convergence < convergenceCriteria){
          converged = 1;
        }

        if(! converged){ //dont overwite last avg vel when done 
           prev_max_movment = max_movment;
        }
      }

      if (!converged){ //dont overwite old perm when done 
        //permeability_old = permeability_new;

        //getTimeInMs(iteration_end);
        time_iteration = (((double) (iteration_end - iteration_start)) / toSec); 
        
        MLUPS = MLN / time_iteration;
        time_lbm += time_iteration;
        
        //calculateAvrageVelocityMPI(&local_avg_velocity,mpi_rank);  
        
        printState();

      }
    }

    /*
    if(!(iter_count % iterSave) && iter_count > 0)
    {	
      //printf("CPU: %f / %e\n", MLN, time_iteration);
      _lbm->printMLUPS(stdout);
    }
    */

    MPI_Bcast(&converged,1,MPI_INT,0,MPI_COMM_WORLD);
    if (!converged){ //dont increas the itteration count when finished
      iter_count++;
    }

  }

  calculateAvrageVelocityMPI(local_avg_velocities,mpi_rank,num_threads);

  //stop  total timer
  getTimeInMs(t2);
  time_total = (double (t2 - total_start) / toSec);


#endif

}

void printIndevidualTimingsMPI(char *stream,int num_threads)
{
#if MODE == mode_mpi || MODE == mode_hybrid
	int offset = 0;
	for(int i = 0; i < num_threads; i++){
		offset = lbm[i]->printTimings(stream,offset);
	}
#endif
}

void copyAvgVelToCPU_MPI(real_t *u_new, int mpi_size,int ant_threads)
{
  //TODO: Validate
#if MODE == mode_mpi || MODE == mode_hybrid   
  int activeElement = 0;
  for(int i = 0; i < ant_threads; i++){
    activeElement += lbm[i]->getActiveElements();
  }

  real_t *local_vel = new real_t[activeElement];
  int offset = 0;
  for(int i = 0; i < ant_threads; i++){
    //activeElement += lbm[i]->getActiveElements();
    offset += lbm[i]->copyAvgVelToCPU(local_vel + offset);
  }


  //real_t *local_vel = new real_t[activeElement];
  
  int *sizes = new int[mpi_size];
  int *offsets = new int[mpi_size];


  MPI_Gather(&activeElement,1,MPI_INT,sizes,1,MPI_INT,0,MPI_COMM_WORLD);

  offsets[0] = 0;
  for(int i = 1; i < mpi_size; i++){
    offsets[i] = offsets[i-1] + sizes[i-1];
  }

  //_lbm->copyAvgVelToCPU(local_vel);
  MPI_Gatherv(local_vel,activeElement,MPI_FLOAT,u_new,sizes,offsets,MPI_FLOAT,0,MPI_COMM_WORLD);

#endif
}


void calculateAvrageVelocityMPI( real_t *avg_velocities, int mpi_rank, int ant_threads)
{
#if MODE == mode_mpi || MODE == mode_hybrid   
  long t1,t2;
  
  for(int i = 0; i < ant_threads; i++){
      lbm[i]->calculateAvrageVelocity(&avg_velocities[i]);
    }

    getTimeInMs(t1);
    //TODO: have to fix this timing
    waitToFinish(lbm,ant_threads,AVG_VEL);
    getTimeInMs(t2);
    time_wait_to_finish += (double (t2 - t1) / toSec);



    real_t local_avg_velocity = 0.0;
    for(int i = 0; i < ant_threads; i++){
      local_avg_velocity += avg_velocities[i];
    }

  //have to do this as MPI_reduce don't work with double when MPI_FLOAT is spesified
  real_t tmp_avg_vel;
  MPI_Reduce(&local_avg_velocity,&tmp_avg_vel,1,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD);
  avg_velocity = (double) tmp_avg_vel;

 if(mpi_rank == 0){
    avg_velocity = (avg_velocity / (double) ActivevoxelCount);
    permeability_new = getPermeability(avg_velocity, true);
 }


#endif



}