#include "Simulation_threaded.h"


//Timings 
extern double time_total;
extern double time_lbm;
extern double time_setup;
extern double time_wait_to_finish;
extern double time_wait_to_finish_comm;
extern double time_reduce;
extern double time_iteration;


extern double avg_velocity;
extern double permeability_old;
extern double permeability_new;

extern int extra;
//extern int size_x;
//extern int size_y;
//extern int size_z; 

extern bool write_to_file;

extern double viscosity;
extern real_t porosity;
//double cellSpacingX;

extern double MLN;

extern int iter_count;
extern uchar *data;
extern long long int ActivevoxelCount;

extern double last_avg_velocity;
extern double convergence;

extern double MLUPS;
extern int *solid_size;
extern int *solid_offset;

extern real_t prev_max_movment;
extern real_t max_movment;
extern double rate;

//used in both run simulation and print timings
extern LBM_3D **lbm;




void runSimulationThreaded(int num_threads)
{

  
  uchar **local_data_ptr;
  int local_size_z;
  CommandQueue *queues;

  real_t *avg_velocities;
  real_t *max_movments;
  real_t omega;


  //Timings
  long t1;
  long total_start;
  long t2;

  long iteration_start;
  long iteration_end;

  
  time_total = 0.0;
  //time_loaddata = 0.0;
  time_lbm = 0.0;
  time_setup = 0.0;
  time_wait_to_finish = 0.0;
  time_wait_to_finish_comm = 0.0;

  int hits = 0;
  avg_velocity = 0.0;
  last_avg_velocity = 0.0;

  int converged = 0;


  //Start timer for setup and the total 
  getTimeInMs(t1);
  getTimeInMs(total_start);

  	
  //initilize openCL posible only do this when GPU mode  
  queues = new CommandQueue[num_threads];
  setupOpenCL(queues,num_threads,0);

  
  
  lbm = new LBM_3D *[num_threads]; 
  avg_velocities = new real_t[num_threads]; 
  max_movments = new real_t[num_threads]; 
  local_data_ptr =  (uchar**) malloc(num_threads * sizeof(uchar*));
  solid_size = new int[num_threads];
  solid_offset = new int [num_threads];

  
  omega = 1.f / tau;
  viscosity = (1./6. * (2./omega - 1.)); 



  
  //First part of setup
  getTimeInMs(t2);
  time_setup = (double (t2 - t1) / toSec);
  //printf("First part setup %ld ms \n",t2 -t1);

  
  //Start timer for second part of setup
  getTimeInMs(t1);
  
  local_size_z =  ceil( (double)getSizeZ() / (double) num_threads);
  cout << "Global stone size " << getSizeX() << " " << getSizeY() << " " << getSizeZ() << endl;
  //cout << "num threads " << num_threads << endl; 
  //cout << "Local size_z " << local_size_z << endl;


  //used to get correct offset when periodic in Z direction
  int ghost_layer_tmp = 0;
  uchar *data_ptr;
#if PERIODIC_Z
  data_ptr = &data[GHOST_LAYERS * getSizeX() * getSizeY()];
  ghost_layer_tmp = GHOST_LAYERS;
#else
  data_ptr = data;
#endif


  solid_offset[0] = 0;
  for(int i = 0; i < num_threads; i++){
    solid_size[i] = (i != (num_threads - 1)) ? local_size_z : (getSizeZ() - solid_offset[i]);
    
    local_data_ptr[i] = &data_ptr[(solid_offset[i] - ghost_layer_tmp) * getSizeX() * getSizeY()];
    
    if (i != num_threads -1){
      solid_offset[i + 1] = solid_offset[i] + solid_size[i];
    }

    ghost_layer_tmp = GHOST_LAYERS;
  }

//cout << "INIT solvers ant " << num_threads << endl;

  //Init the LBM solvers
  for(int i = 0; i < num_threads; i++){
    //cout << "inni for " << endl;
    int ghost_layers_north = 0;
    int ghost_layers_south = 0;


    if (!PERIODIC_Z && i == 0){
      ghost_layers_north = 0;
    }else{
      ghost_layers_north = GHOST_LAYERS;
    }

    if (! PERIODIC_Z && i == num_threads -1){
      ghost_layers_south = 0;
    }else{
      ghost_layers_south = GHOST_LAYERS;
    }

    //lbm[i] = new LBM_3D(i,local_data_ptr[i], omega, getSizeX(), getSizeY(), solid_size[i], extra,
    //ghost_layers_north, ghost_layers_south, write_to_file, queues[i],force);
    //cout << "Creating solver  " << i << endl;
    lbm[i] = new LBM_3D(i,local_data_ptr[i], omega, getSizeX(), getSizeY(), solid_size[i], extra,
      ghost_layers_north, ghost_layers_south, write_to_file, queues[i],force);
  }

  //Have to set the north and south border after creation all all 
  for(int i = 0; i < num_threads; i++){

    LBM_3D *north_neighbour;
    LBM_3D *south_neighbour;
    if (!PERIODIC_Z && i == 0){
    north_neighbour = NULL;
    }else{
    north_neighbour = lbm[(i + num_threads -1) % num_threads];
    }

    if (!PERIODIC_Z && i == num_threads -1){
    south_neighbour = NULL;
    }else{
    south_neighbour = lbm[(i+1) % num_threads];
    } 

    lbm[i]->setNorthNeighbour(north_neighbour, -1);
    lbm[i]->setSouthNeighbour(south_neighbour, -1);

  }

  //second  part of setup
  getTimeInMs(t2);
  time_setup += (double (t2 - t1) / toSec);
  //printf("second part setup %ld ms \n",t2 -t1);


  while (iter_count < iterMax  && !converged){
    //start timer for this iteration
    getTimeInMs(iteration_start);
      
    //Exchange layer before doing itterqation to ernshire correct on first run allso
    //Copy ghost layers to CPU
    for(int i = 0; i < num_threads; i++){
      lbm[i]->copyGhostLayersToCPU();
    }

    getTimeInMs(t1);
    waitToFinish(lbm,num_threads,COPY_TO_CPU);
    getTimeInMs(t2);
    time_wait_to_finish_comm += (double (t2 - t1) / toSec);

	
	for(int i = 0; i < num_threads; i++){
		lbm[i]->doCollide_inner();
	}

    //Exchange ghost layers
    for(int i = 0; i < num_threads; i++){
      lbm[i]->exhangeGhostLayers_threaded();
    }



    //Copy ghost layers to GPU
    for(int i = 0; i < num_threads; i++){
      lbm[i]->copyGhostLayersToGPU();
    }

    getTimeInMs(t1);
    waitToFinish(lbm,num_threads,COPY_TO_GPU);
    getTimeInMs(t2);
    time_wait_to_finish_comm += (double (t2 - t1) / toSec);



    //Collide and stream step
//#pragma omp parallel for
    for(int i = 0; i < num_threads; i++){
      lbm[i]->doCollide_ghost();
      lbm[i]->doStream();
    }

    getTimeInMs(t1);
    waitToFinish(lbm,num_threads,COLLIDE|STREAM);
    getTimeInMs(t2);
    time_wait_to_finish += (double (t2 - t1) / toSec);

    getTimeInMs(iteration_end);

   

    time_iteration = (((double) (iteration_end - iteration_start)) / toSec);
    for(int i = 0; i < num_threads; i++){
      lbm[i]->total_lbm_wall_time += time_iteration;
    }
 /*  
    
    //Calculate avg velocity
    for(int i = 0; i < num_threads; i++){
      lbm[i]->calculateAvrageVelocity(&avg_velocities[i]);
    }

    getTimeInMs(t1);
    //TODO: have to fix this timing
    waitToFinish(lbm,num_threads,AVG_VEL);
    getTimeInMs(t2);
    time_wait_to_finish += (double (t2 - t1) / toSec);



    //avg_velocity = 0.0;
    real_t tmp_avg_vel = 0.0;
    for(int i = 0; i < num_threads; i++){
      //printf("lbm %d %e \n",i,avg_velocities[i]);
      //avg_velocity += avg_velocities[i];
      tmp_avg_vel += avg_velocities[i];
      //lbm[i]->writeVelocityFieldToFile();
    }
    
    //exit(-1);
    //printf("%d Avg vel: %e\n",iter_count,avg_velocity);
    avg_velocity = (double) tmp_avg_vel;
    avg_velocity = (avg_velocity / (double) ActivevoxelCount);
*/

    if(!(iter_count % iterSave))
    {

      max_movment = 0.0;
      for(int i = 0; i < num_threads; i++){
        lbm[i]->getMaxParticleMovment(&max_movments[i]);
        max_movment = max(max_movment, max_movments[i]);
      }

      rate =  (prev_max_movment / max_movment);
      convergence = fabs(1.0 - rate); //(prev_max_movment / max_movment));

      //printf("Converged: %d %e prev %e max %e rate %e  \n", iter_count, convergence,prev_max_movment, max_movment, rate);

      


      if(convergence < convergenceCriteria){
        converged = 1;
      }else{
        prev_max_movment = max_movment;
      }
    }

    if(!converged){ //dont overwite last avg vel when done 
      //getTimeInMs(iteration_end);
      //time_iteration = (((double) (iteration_end - iteration_start)) / toSec); 
      MLUPS = MLN / time_iteration;

      time_lbm += time_iteration;
    
      //calculateAvrageVelocityThreaded(avg_velocities,num_threads);

      printState();
      iter_count++;
    }

  }
  printf("Done simulating\n");
  calculateAvrageVelocityThreaded(avg_velocities,num_threads);
 


 
  //stop  total timer
  getTimeInMs(t2);
  time_total = (double (t2 - total_start) / toSec);

  /*
  delete[] avg_velocities;
  delete[] lbm;
  delete[] max_movments;
  free(local_data_ptr);
  delete[] solid_size;
  delete[] solid_offset;
*/
}

void printIndevidualTimingsThreaded(int ant_threads)
{
	char *buf = (char *) malloc(1000 * sizeof(char));
  for(int i = 0; i < ant_threads; i++){
      //lbm[i]->printTimings(stdout);
	  lbm[i]->printTimings(buf,0);
	  printf("%s",buf);
  }
  free(buf);
}

void copyAvgVelToCPU_threaded(real_t *u_new,int ant_threads)
{
  int offset = 0;
  for(int i = 0; i < ant_threads; i++){
    offset += lbm[i]->copyAvgVelToCPU(u_new + offset);
  }
}

void calculateAvrageVelocityThreaded(real_t *avg_velocities, int ant_threads)
{
    long t1,t2;
    //Calculate the avrage velocity as the simulation is compleeted
    for(int i = 0; i < ant_threads; i++){
      lbm[i]->calculateAvrageVelocity(&avg_velocities[i]);
    }

    getTimeInMs(t1);
    //TODO: have to fix this timing
    waitToFinish(lbm,ant_threads,AVG_VEL);
    getTimeInMs(t2);
    time_wait_to_finish += (double (t2 - t1) / toSec);



    real_t tmp_avg_vel = 0.0;
    for(int i = 0; i < ant_threads; i++){
      tmp_avg_vel += avg_velocities[i];
    }

    avg_velocity = (double) tmp_avg_vel;
    avg_velocity = (avg_velocity / (double) ActivevoxelCount);

    permeability_new = getPermeability(avg_velocity, true);

}
