#include "Simulation_utils.h"





int extra  = 0;
int size_x = 0;
int size_y = 0;
int size_z = 0; 
bool write_to_file;
int b_x;
int b_y;
int b_z;

uchar *data;

double cellSpacingX;

double viscosity;
real_t porosity;

long long int voxelCount;
long long int ActivevoxelCount;


double avg_velocity;
double last_avg_velocity;

double permeability_old;
double permeability_new;
double convergence;


int iter_count = 0;

//Timings:
double time_total = 0.0;
double time_loaddata = 0.0;
double time_lbm = 0.0;
double time_setup = 0.0;
double time_wait_to_finish = 0.0;
double time_wait_to_finish_comm = 0.0;
double time_reduce = 0.0;
double time_iteration = 0.0;

double MLN;
double MLUPS;


int *solid_size;
int *solid_offset;

LBM_3D **lbm;

real_t prev_max_movment;
real_t max_movment;
double rate;

void createLattice(int d_x, int d_y, int d_z, int d_extra)
{
	size_x = d_x+d_extra;
	size_y = d_y;
	size_z = d_z;
	extra = d_extra;
  data = (uchar *) malloc (size_x*size_y*size_z*sizeof(uchar));

	for(int z=0; z<size_z; z++)
	{
		for(int y=0; y<size_y; y++)	
		{
			for(int x=0; x<size_x; x++)
			{
				int i = x + y*size_x + z*size_x*size_y;

				data[i] = 0; //0.0f;	
			}
		}
	}
}


void createSolidBoundary(int solid_x, int solid_y, int solid_z)
{	
  if(solid_x == 1)
		b_x = 2;

	if(solid_y == 1)
		b_y = 2;

	if(solid_z == 1)
		b_z = 2;

	for(int z=0; z<size_z; z++)
	{
		for(int y=0; y<size_y; y++)	
		{
			for(int x=0; x<size_x; x++)	
			//for(int x=(extra/2); x<size_x-(extra/2); x++)
			{
				int i = x + size_x*y + z * size_x*size_y;
	
				if(solid_x == 1)
				{
          if(x == 0 || x == size_x-1){
							data[i] = 2;
          }
				} 

				if(solid_y == 1)
				{
          if(y == 0 || y == size_y-1){	
							data[i] = 2;
          }
				}
		
				if(solid_z == 1)
				{
          if(z == 0 || z == size_z-1){
							data[i] = 2;
          }
				}
			}
		}
	}
}

void loadData(int extra)
{
	write_to_file = true;

  int s_x, s_y, s_z;
  uchar *d;

  double startX;
  double startY;
  double startZ;

  double endX;
  double endY;
  double endZ;
 	
	
	ifstream mineralFile(DATA_FILE);

	
	if (!mineralFile.is_open()) 
	{
		printf("Error: Unable to open mineral file\n");
    exit(-1);
	}

	// Skipp first 2 lines
	std::string fileLine;
	getline(mineralFile,fileLine);
	getline(mineralFile,fileLine);

	// Read size of the grid
	mineralFile >> s_x >> s_y >> s_z;
	
	createLattice(s_x, s_y+2, s_z+2, extra);

	d = (uchar *) malloc (s_x*s_y*s_z*sizeof(uchar));

  //get information about the physical size
  mineralFile >> startX >> endX  >> startY >> endY >> startZ >> endZ;
  cellSpacingX = (endX - startX) / (double) s_x;

  printf("cellSpacing X %f\n", cellSpacingX);
  
  getline(mineralFile,fileLine);
  //getline(mineralFile,fileLine);
	//cout << fileLine << endl;

	int mineral;

   voxelCount = -1;
   //voxelCount = 0;
   
   ActivevoxelCount = 0;

	// ofstream ofs("grid.bin", ios_base::binary);

	while(mineralFile >> mineral) 
	{
		voxelCount++; 

		mineral *= 3;
    if (mineral == 0){
			mineral = 0;
      ActivevoxelCount ++;
    }
		else if(mineral == 3)
			mineral = 3;

		d[voxelCount] = mineral;
		
		//ofs.put(static_cast<char>(mineral));
	}

	for(int z=0; z<s_z; z++)
	{
		for(int y=0; y<s_y; y++)	
		{
			for(int x=0; x<s_x; x++)
			{
				int i_a = x + s_x*y + z*s_x*s_y;

				int i_b = (x+(extra/2)) + size_x*(y+1) + (z+1)*size_x*size_y;

				data[i_b] = d[i_a];
			}
		}
	}

	createSolidBoundary(0, 1, 1);
  

	mineralFile.close();
  //free(d);
}

void extendData(int layers_north, int layers_south)
{

  int new_size_z = getSizeZ() + layers_north + layers_south;

  uchar *data_new;
  data_new = (uchar *) malloc(getSizeX() * getSizeY() * new_size_z * sizeof(uchar));

  //Copy layers_north of the 'bottom' layers to the top
  memcpy(data_new,data +(getSizeX() * getSizeY() * (getSizeZ() - layers_north)),getSizeX() * getSizeY() * layers_north*sizeof(uchar));
  //Copy the datablok over
  memcpy(data_new +(getSizeX() * getSizeY() * layers_north),data,getSizeX() * getSizeY() * getSizeZ()*sizeof(uchar));
  //Copy layers_south of the 'top' layers to the bottom
  memcpy(data_new +(getSizeX() * getSizeY() * (layers_north+getSizeZ())),data, getSizeX() * getSizeY() * layers_south*sizeof(uchar));

  //free the old data
  free(data);

 data = data_new;
}

//void createPoiseuilleFlow1D(int L)
void createPoiseuilleFlow1D(int x, int y, int z)
{
	write_to_file = true;

	createLattice(x, y, z, 0);

  //Calculation off this is done when reading the stone, but as the stone is never read we calculate it
  //the to and bottom layer is solid so does not cout to the active voxels
  ActivevoxelCount = x * (y-2) * z; 
  //Also have to set the cellspacing
  cellSpacingX = 1.0;
	createSolidBoundary(0, 1, 0);
}

void createPoiseuilleFlow2D(int L)
{
	write_to_file = true;

	createLattice(L, L, L, 0);
  //Calculation off this is done when reading the stone, but as the stone is never read we calculate it
  //the to and bottom layer is solid so does not cout to the active voxels
  ActivevoxelCount = L * (L-2) * (L-2); 
	createSolidBoundary(0, 1, 1);
}

void createBenchmark(int L)
{
	write_to_file = true;

	createLattice(L, L, L, 0);
}


int getSizeX(){
  return size_x;
}
int getSizeY(){
  return size_y;
}
int getSizeZ(){
  return size_z;
}

int getIndex (int x, int y, int z){
  return x + y*getSizeX() + z*getSizeX()*getSizeY();
}

uchar getSolid(int x,int y,int z){
  return data[x + y*getSizeX() + z*getSizeX()*getSizeY()];
}



real_t getPorosity()
{
	double fluid = 0;

	for( int z=0; z<getSizeZ(); z++)
	{
		for( int y=0; y<getSizeY(); y++)
		{
			for( int x=(extra/2); x<getSizeX()-(extra/2); x++)//FIXME: was this
			//for( int x=0; x<getSizeX(); x++)

			{	
				if(getSolid(x,y,z) == 0)
					fluid++;
			}
		}
	}
	//return fluid /((getSizeX()-boundaries[0])*(getSizeY()-boundaries[1])*(getSizeZ()-boundaries[2])); FIXME: was this but not in use
  //return fluid /((getSizeX()-boundaries[0]-extra )*(getSizeY()-boundaries[1])*(getSizeZ()-boundaries[2]));
  return fluid /((getSizeX()-b_x-extra )*(getSizeY()-b_y)*(getSizeZ()-b_z));
  //return voxelCount /((getSizeX()-b_x-extra )*(getSizeY()-b_y)*(getSizeZ()-b_z));

}


double getConvergence()
{
	return fabs(avg_velocity - last_avg_velocity);
	//return (avg_velocity - last_avg_velocity);
}

double getPermeability(double averageVelocity, bool doDeltaF)
{	
  double constant = (viscosity * cellSpacingX * cellSpacingX * toMilliDarcy * avrage_dencity); 
  constant = constant / force;

	if(doDeltaF) 
	{
		return averageVelocity * porosity * constant; 
	} 
	else 
	{
		return averageVelocity * porosity * constant;
	}
}

void waitToFinish(LBM_3D ** lbm, int ant, int place)
{
   for(int i = 0; i < ant; i++){
    lbm[i]->finish(place);
   }
}


void printState()
{
	if(!(iter_count % iterSave) && iter_count > 0)
	{	
    printf("%d (%ds): ", iter_count, (int) time_lbm);
    //printf("AvgVel: %e, ", avg_velocity);
	printf("Convergence: %e, ", convergence);
    //printf("Perm: %lf, ", permeability_new);
    printf("MLUPS: %f (%e sek)", MLUPS, time_iteration);
    printf("\n");
		

    std::string dummy;
    std::string file_n_convergence;
    file_n_convergence = LOGFOLDER + dummy + "conv.txt";

    FILE *p_convergence;
    p_convergence = fopen(file_n_convergence.c_str(),"a+");
    fprintf(p_convergence, "%d %e %e \n",iter_count, rate, convergence ); 
    fclose(p_convergence);

    /*

		
		printf("Perm: %lf ", permeability_new);
		printf("LBM: %fls ", time_lbm);
		printf("Collide: %fs ", time_collide);
		printf("Stream: %fs ", time_stream);	
		printf("Lines: %fs ", time_lines);
		printf("Avg Vel: %fs \n", time_avg_velocity);
		*/
    
    /*
		// Write simulation to file
		fprintf(pFile, "%d (%ds) ", iter_count);
    fprintf(pFile, "AvgVel: %e ", avg_velocity);
		fprintf(pFile, "Convergence: %e ", convergence);
    fprintf(pFile, "Perm: %lf ", permeability_new);
		fprintf(pFile, "\n");
    */
    
    /*
    fprintf(pFile, "MLUPS: %f (%f) ", MLUPS, best_mlups);		
		fprintf(pFile, "LBM: %fs ", time_lbm);
		fprintf(pFile, "Collide: %fs ", time_collide);
		fprintf(pFile, "Stream: %fs ", time_stream);	
		fprintf(pFile, "Lines: %fs ", time_lines);
		fprintf(pFile, "Avg Vel: %fs \n", time_avg_velocity);
		*/
	}
/*
	FILE *p_time;
	p_time = fopen(file_n_time.c_str(), "a+");
	fprintf(p_time, "%f %f %f \n", time_lbm, time_collide, time_stream);
	fclose(p_time);

	FILE *p_permeability;
	p_permeability = fopen(file_n_permeability.c_str(),"a+");
	fprintf(p_permeability, "%1.16f \n", permeability_new);
	fclose(p_permeability);

	FILE *p_convergence;
	p_convergence = fopen(file_n_convergence.c_str(),"a+");
	fprintf(p_convergence, "%1.16f \n", convergence);
	fclose(p_convergence);

	FILE *p_mlups;
	p_mlups = fopen(file_n_MLUPS.c_str(),"a+");
	fprintf(p_mlups, "%f \n", MLUPS);
	fclose(p_mlups);

	FILE *p_vel;
	p_vel = fopen(file_n_vel.c_str(),"a+");
	fprintf(p_vel, "%e \n", u_max);
	fclose(p_vel);
*/
}

void printFinaleState( FILE *stream )
{
  //Print the final state 

  if (iter_count < iterMax ){
    fprintf(stream, "\n*************\nSystem convergered after %d iterations.\nFinal results:\n",iter_count);
  }else{
    fprintf(stream, "\n*************\nSystem DID NOT convergered after %d iterations.\nFinal results:\n",iterMax);
  }



  fprintf(stream, "Kinematic visc.   = %f\n", viscosity); //0.0555556
  fprintf(stream, "Average velocity  = %e\n", avg_velocity); //9.07042e-05
  fprintf(stream, "Reynolds number   = %f\n", avg_velocity/viscosity); //0.00163268
  fprintf(stream, "\n");
  fprintf(stream, "Porosity          = %f\n", porosity); //0.389336
  //printf("Perm_X_old (mD)   = %f\n", permeability_old); //59637.2
  fprintf(stream, "Perm_X_new (mD)   = %f\n", permeability_new); //59637.2

}

/*
 * Have iter_count and viscosity this shoud maybe be a parameter 
 */
void writePoiseuilleFlow2DToFile(FILE *pFile, int *h_solids, real_t *u_new)
{
  //printf("writePoiseuilleFlow2DToFile: (%d %d %d)\n",getSizeX(), getSizeY(), getSizeZ());
  //FILE *pFile;
  //pFile = fopen("logs\\right_1d_profile.txt","a+");
  fprintf(pFile,"Iteration: %d \n", iter_count);
  fprintf(pFile,"Viscosity: %f \n", viscosity);
  //fprintf(pFile,"Max Velocity: %f \n", u_max);

  real_t r = (( (real_t)getSizeY()) - 2 ) / 2.;
  real_t g = (force/(2*viscosity*avrage_dencity));

  double total_error = 0.0f;

  for(int iY=1; iY<getSizeY()-1; iY++)
  {		
    double correct = g * (pow(r, 2) - pow(((real_t) iY - 0.5 - r), 2));

    //real_t our = u_new[getIndex(getSizeX()/2, iY, (getSizeZ()-1)/2)];
    int index = h_solids[getIndex(getSizeX()/2, iY, getSizeZ()/2)];
    real_t our = u_new[index];
    double real_error = (correct  - our);
    total_error += pow(correct  - our, 2) / ((real_t) getSizeY()-2);

    fprintf(pFile, "%d %f %f %e %e \n",iY, our, correct, real_error,our/correct );       
  }

  fprintf(pFile,"Error: %f\n", sqrtf(total_error));
  fprintf(pFile,"Error: %6.16f\n", sqrtf(total_error));
  fprintf(pFile,"Error: %e\n", sqrtf(total_error));
  fprintf(pFile,"\n");
  //fclose(pFile);

  printf("Error: %e\n", sqrtf(total_error));
}