//#pragma OPENCL EXTENSION cl_amd_printf : enable
//#pragma OPENCL EXTENSION cl_khr_fp64: enable

typedef float real_t;
//typedef double real_t;
//#define real_t double
#ifdef WIN32
#define CONST __constant
#else
#define CONST const
#endif


CONST sampler_t sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_NONE | CLK_FILTER_NEAREST;

__constant real_t t_3d[19] = 
{ 
	1.f/3.f,  // 0
	1.f/18.f, // 1
	1.f/18.f, // 2
	1.f/18.f, // 3
	1.f/36.f, // 4
	1.f/36.f, // 5
	1.f/36.f, // 6
	1.f/36.f, // 7
	1.f/36.f, // 8
	1.f/36.f, // 9      
	1.f/18.f, // 10
	1.f/18.f, // 11
	1.f/18.f, // 12         
	1.f/36.f, // 13
	1.f/36.f, // 14
	1.f/36.f, // 15   
	1.f/36.f, // 16
	1.f/36.f, // 17
	1.f/36.f  // 18
};

__constant real_t c_3d[19*3] = 
{
	0, 0, 0,    // 0
	-1, 0, 0,   // 1
	0, -1, 0,   // 2
	0, 0, -1,   // 3
	-1, -1, 0,  // 4
	-1, 1, 0,   // 5
	-1, 0, -1,  // 6 
	-1, 0, 1,   // 7 
	0, -1, -1,  // 8
	0, -1, 1,   // 9
	1, 0, 0,    // 10
	0, 1, 0,    // 11
	0, 0, 1,    // 12
	1, 1, 0,    // 13
	1, -1, 0,   // 14
	1, 0, 1,    // 15
	1, 0, -1,   // 16
	0, 1, 1,    // 17
	0, 1, -1    // 18
};


////////////////////////////////////////////////////////////////////////////////
//Util functions
////////////////////////////////////////////////////////////////////////////////

int get_next_x (int x, int size_x, int f_index, bool periodic)
{
	int nextX = x + c_3d[f_index*3 + 0];

	if(periodic)
	{
		if(nextX == -1)
			nextX = size_x-1;

		if(nextX == size_x)
			nextX = 0;
	}
	
	return nextX;
}

int get_next_y (int y, int size_y, int f_index, bool periodic)
{
	int nextY = y + c_3d[f_index *3 + 1];

	if(periodic)
	{
		if(nextY == -1)
			nextY = size_y-1;

		if(nextY == size_y)
			nextY = 0;
	}

return nextY;
}

int get_next_z (int z, int size_z, int f_index, bool periodic)
{
	int nextZ = z + c_3d[f_index *3 + 2];

	if(periodic)
	{
		if(nextZ == -1)
			nextZ = size_z-1;

		if(nextZ == size_z)	
			nextZ = 0;
	}

	return nextZ;
} 

void swap_global (__global real_t *t_1, __global real_t *t_2, real_t *current_max)
{

  real_t tmp = *t_1;
  *t_1 = *t_2;
  *t_2 = tmp;
  *current_max = max(*current_max,fabs(*t_1 - tmp));
}

void copy_to_shared(__global real_t *f0,
                     __global real_t *f1,
                     __global real_t *f2,
                     __global real_t *f3,
                     __global real_t *f4,
                     __global real_t *f5,
                     __global real_t *f6,
                     __global real_t *f7,
                     __global real_t *f8,
                     __global real_t *f9,
                     __global real_t *f10,
                     __global real_t *f11,
                     __global real_t *f12,
                     __global real_t *f13,
                     __global real_t *f14,
                     __global real_t *f15,
                     __global real_t *f16,
                     __global real_t *f17,
                     __global real_t *f18,
                     __local real_t *lattice,
                     int global_index)
{

  lattice[ 0] = f0[global_index];
  lattice[ 1] = f1[global_index];
  lattice[ 2] = f2[global_index];
  lattice[ 3] = f3[global_index];
  lattice[ 4] = f4[global_index];
  lattice[ 5] = f5[global_index];
  lattice[ 6] = f6[global_index];
  lattice[ 7] = f7[global_index];
  lattice[ 8] = f8[global_index];
  lattice[ 9] = f9[global_index];
  lattice[10] = f10[global_index];
  lattice[11] = f11[global_index];
  lattice[12] = f12[global_index];
  lattice[13] = f13[global_index];
  lattice[14] = f14[global_index];
  lattice[15] = f15[global_index];
  lattice[16] = f16[global_index];
  lattice[17] = f17[global_index];
  lattice[18] = f18[global_index];
}

void copy_to_global(__global real_t *f0,
                     __global real_t *f1,
                     __global real_t *f2,
                     __global real_t *f3,
                     __global real_t *f4,
                     __global real_t *f5,
                     __global real_t *f6,
                     __global real_t *f7,
                     __global real_t *f8,
                     __global real_t *f9,
                     __global real_t *f10,
                     __global real_t *f11,
                     __global real_t *f12,
                     __global real_t *f13,
                     __global real_t *f14,
                     __global real_t *f15,
                     __global real_t *f16,
                     __global real_t *f17,
                     __global real_t *f18,
                     __local real_t *lattice,
                     int global_index)
{

  f0[global_index ] = lattice[ 0];
  f1[global_index ] = lattice[ 1];
  f2[global_index ] = lattice[ 2];
  f3[global_index ] = lattice[ 3];
  f4[global_index ] = lattice[ 4];
  f5[global_index ] = lattice[ 5];
  f6[global_index ] = lattice[ 6];
  f7[global_index ] = lattice[ 7];
  f8[global_index ] = lattice[ 8];
  f9[global_index ] = lattice[ 9];
  f10[global_index] = lattice[10];
  f11[global_index] = lattice[11];
  f12[global_index] = lattice[12];
  f13[global_index] = lattice[13];
  f14[global_index] = lattice[14];
  f15[global_index] = lattice[15];
  f16[global_index] = lattice[16];
  f17[global_index] = lattice[17];
  f18[global_index] = lattice[18];
}


inline real_t get_rho_shared(__local real_t *lattice)
{
  real_t tmp1 = lattice[0]  + lattice[1] + lattice[2] + lattice[3] + lattice[4] + 
    lattice[5] + lattice[6] + lattice[7] + lattice[8] + lattice[9],
         tmp2 = lattice[10] + lattice[11] + lattice[12] + lattice[13] + lattice[14] + 
    lattice[15] + lattice[16] + lattice[17] + lattice[18];
  return tmp1+tmp2;
}

inline real_t get_ux_shared(__local real_t *lattice, real_t rho)
{
	return ((lattice[13] - lattice[4]) + 
	        (lattice[14] - lattice[5]) + 
	        (lattice[15] - lattice[6]) + 
	        (lattice[16] - lattice[7]) + 
	        (lattice[10] - lattice[1])) / rho;
}

inline real_t get_uy_shared(__local real_t *lattice, real_t rho)
{
   	return ((lattice[5] - lattice[14]) +  
   	        (lattice[11] - lattice[2]) +  
   	        (lattice[13] - lattice[4]) + 
   	        (lattice[17] - lattice[8]) + 
   	        (lattice[18] - lattice[9])) / rho;
}		

inline real_t get_uz_shared(__local real_t *lattice, real_t rho)
{
	 return ((lattice[7] - lattice[16]) +  
	         (lattice[9] - lattice[18]) + 
	         (lattice[15] - lattice[6]) +  
	         (lattice[17] - lattice[8]) + 
	         (lattice[12] - lattice[3])) / rho;			
}		
	





////////////////////////////////////////////////////////////////////////////////
// Colide and swap
////////////////////////////////////////////////////////////////////////////////
__kernel void collide_and_swap(__read_only image3d_t solids,
                              __global real_t *f0,
                              __global real_t *f1,
                              __global real_t *f2,
                              __global real_t *f3,
                              __global real_t *f4,
                              __global real_t *f5,
                              __global real_t *f6,
                              __global real_t *f7,
                              __global real_t *f8,
                              __global real_t *f9,
                              __global real_t *f10,
                              __global real_t *f11,
                              __global real_t *f12,
                              __global real_t *f13,
                              __global real_t *f14,
                              __global real_t *f15,
                              __global real_t *f16,
                              __global real_t *f17,
                              __global real_t *f18,
                              __local real_t *lattice,
                              real_t omega,
                              real_t force,
                              real_t avrage_dencity,
                              int ant_vec,
							  int offset_x,
							  int offset_y,
							  int offset_z ) 
{

	const int x = get_global_id(0) + offset_x;
	const int y = get_global_id(1) + offset_y;
	const int z = get_global_id(2) + offset_z;

	const int nx = get_global_size(0);
	const int ny = get_global_size(1);
	const int nz = get_global_size(2);

	//const int lx = get_local_id(0);
  const int lnx = get_local_size(0);
  const int lny = get_local_size(1);
  const int lnz = get_local_size(2);

	const int lx = (get_local_id(2) * lny * lnx) + (get_local_id(1) * lnx) +  get_local_id(0);

	int4 this_coord = (int4) (x,y,z,0);
	int this_index =  read_imagei(solids,sampler,this_coord).x;



	//const int global_index = (nx * ny * z) + (nx * y) + (x);
	const int local_index = lx * ant_vec;

	//Make pointer to this threads starting pos in the shared lattice memory
	__local real_t *local_lattice = &lattice[local_index];

	if (this_index != -1) { 
		copy_to_shared(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,
			f11,f12,f13,f14,f15,f16,f17,f18,
			local_lattice,this_index);
	}
	barrier(CLK_LOCAL_MEM_FENCE);


	if (this_index == -1) {
		return;
	}


	// Compute density from the particle distribution functions 
	real_t delta_rho = get_rho_shared(local_lattice);
	real_t rho = delta_rho + avrage_dencity;  // flyttalls triks
	//real_t rho = delta_rho;


	//Compute velocity in x, y, and z directions from the particle distribution functions			
	real_t ux = get_ux_shared(local_lattice,rho);
	real_t uy = get_uy_shared(local_lattice,rho); 
	real_t uz = get_uz_shared(local_lattice,rho);


	//precalculeting some of the constants used later
	real_t uSqr = ux*ux + uy*uy + uz*uz;
	//real_t u_sqr = 1.5f * uSqr;
	real_t u_sqr = (1.0f/3.0f) * uSqr;

	//real_t tmp = 3.0f * force / rho;
	real_t f_eq;
	real_t cu;
	real_t external_force = 0.0f;

	//printf("(%d, %d , %d), rho %e \n",x,y,z,rho);



  
  // Because of the trick for incresing the precission of float f_eq = h_eq
  // the term ((c_3d[0][0] - ux) * (f_eq + t_3d[0]*avrage_dencity)) * tmp; expresses the external force
  // The external force is due to eq 4.1 in Erik's report

  #define C_3D(i,j) c_3d[3*(i)+(j)]
    
  //f0 - f18
  for(int i = 0; i < 19 ; i++){
    
    //Compute local equalibrium
    cu = C_3D(i,0)*ux + C_3D(i,1)*uy + C_3D(i,2)*uz;

    //f_eq = t_3d[i]*delta_rho + t_3d[i]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); //Flyttalls triks
    //f_eq = t_3d[i]*delta_rho + t_3d[i]*rho*((3.f*cu) + (cu*cu) - (1.0f/3.0f) * uSqr);
    f_eq = t_3d[i]*delta_rho + t_3d[i]*rho*((3.f*cu) + (cu*cu) - u_sqr);
    //f_eq =  delta_rho * t_3d[i]  * ( 1.0f + (3.f*cu) + (4.5f*cu*cu) - u_sqr);


    //external_force = C_3D(i,0) * (19.0f/3.0f) * force * t_3d[i];
    //external_force = C_3D(i,0) * (19.0f/6.0f) * force * t_3d[i]; //Flyttall 
    
    //3.0f stammer fra hastighetsleddet i collisjonsopperetoren (f_eq), fordi vi bruker D3Q19
    external_force = C_3D(i,0) * 3.0f * force * t_3d[i] * avrage_dencity; //Flyttall 
    //external_force = (((C_3D(i, 0) - ux) * (f_eq + t_3d[i]*avrage_dencity)) * tmp); // Erik

    local_lattice[i] += omega * (f_eq - local_lattice[i]) + external_force;
/*
    if (i == 0){
      external_force = 0.0; //3.0 * force * t_3d[i];
      //external_force = t_3d[i] * force;
    }else{
      //external_force = t_3d[i] * C_3D(i,0) * force; 
      external_force = C_3D(i,0) * 3.0 * force * t_3d[i]; //(4.0f / 19.f); 
    }
   
    //external_force = (((C_3D(i, 0) - ux) * (f_eq + t_3d[i]*avrage_dencity)) * tmp); // "Virker"
  
    
    local_lattice[i] += omega * (f_eq - local_lattice[i]) + external_force;
*/
    //Relaxation of the particles distributions
    //local_lattice[i] += omega * (f_eq - local_lattice[i]) + ((C_3D(i, 0) - ux) * (f_eq + t_3d[i]*avrage_dencity)) * tmp; // "Virker"
    //local_lattice[i] += omega * (f_eq - local_lattice[i]) + (C_3D(i, 0) * t_3d[i] * avrage_dencity * tmp);

/*
    cu = C_3D(i,0)*ux + C_3D(i,1)*uy + C_3D(i,2)*uz;
    f_eq = t_3d[i] * ( delta_rho + rho * ((3.f*cu) + (4.5f*cu*cu) - (1.5*uSqr)) );
    local_lattice[i] +=
      - omega * local_lattice[i]
      + f_eq * ( omega - ux * (3.0*force/rho) + C_3D(i,0) * (3.0*force/rho) )
      + t_3d[i]*avrage_dencity * C_3D(i,0) * (3.0*force/rho)
      - t_3d[i]*avrage_dencity * ux * (3.0*force/rho);
  */    
  }
#undef C_3D


/*
  real_t rho_after = get_rho_shared(local_lattice);
  if( fabs(rho_after - delta_rho) > 1.e-7 )
  {
    printf("%d %d %d diff %e\n",x,y,z,rho_after - delta_rho);
  }
*/

  barrier(CLK_LOCAL_MEM_FENCE);

  //swap when coppying back to global memory
  copy_to_global(f0,f10,f11,f12,f13,f14,f15,f16,f17,
                 f18,f1,f2,f3,f4,f5,f6,f7,f8,f9,
                 local_lattice,this_index);

  /*
  printf("(%d,%d,%d) ", x,y,z);
  printf("%e ",f0[this_index]);
  printf("%e ",f10[this_index]);
  printf("%e ",f11[this_index]);
  printf("%e ",f12[this_index]);
  printf("%e ",f13[this_index]);
  printf("%e ",f14[this_index]);
  printf("%e ",f15[this_index]);
  printf("%e ",f16[this_index]);
  printf("%e ",f17[this_index]);
  printf("%e ",f18[this_index]);
  printf("%e ",f1[this_index]);
  printf("%e ",f2[this_index]);
  printf("%e ",f3[this_index]);
  printf("%e ",f4[this_index]);
  printf("%e ",f5[this_index]);
  printf("%e ",f6[this_index]);
  printf("%e ",f7[this_index]);
  printf("%e ",f8[this_index]);
  printf("%e ",f9[this_index]);
  printf("\n");
 */
}


////////////////////////////////////////////////////////////////////////////////
// stream and swap
////////////////////////////////////////////////////////////////////////////////
__kernel void stream_and_swap(__read_only image3d_t solids,
                                __global real_t *f0,
                                __global real_t *f1,
                                __global real_t *f2,
                                __global real_t *f3,
                                __global real_t *f4,
                                __global real_t *f5,
                                __global real_t *f6,
                                __global real_t *f7,
                                __global real_t *f8,
                                __global real_t *f9,
                                __global real_t *f10,
                                __global real_t *f11,
                                __global real_t *f12,
                                __global real_t *f13,
                                __global real_t *f14,
                                __global real_t *f15,
                                __global real_t *f16,
                                __global real_t *f17,
                                __global real_t *f18,
                                __global real_t *max_change)                              
                             
{	
//Used to get the index in the force arrays by a int 4 coord nx ny nz contains ther zize in x y ans z dementtion
//#define GET_INDEX(coord) ((nx * ny * coord.z) + (nx * coord.y) + (coord.x))  
  
  const int x = get_global_id(0);
	const int y = get_global_id(1);
	const int z = get_global_id(2);

  const int nx = get_global_size(0);
	const int ny = get_global_size(1);
	const int nz = get_global_size(2);


  //const int global_index = (nx * ny * z) + (nx * y) + (x);


  int4 this_coord = (int4) (x,y,z,0);
  int4 neighbour_coord;
  real_t _max = 0.0;

  int this_index =  read_imagei(solids,sampler,this_coord).x;
  int neighbour_index;

  if (this_index == -1) {
    return;
  }


#define periodicX true
#define periodicY true
#define periodicZ false
//#define periodicZ true


  //Do not check that not outside since pereodic in all directions
  
  // Swap 1 and 10
  neighbour_coord = (int4)(get_next_x(x,nx,1,periodicX),
                            get_next_y(y,ny,1,periodicY),
                            get_next_z(z,nz,1,periodicZ),
                            0);
  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f10 in %d with f1 in %d \n",this_index, neighbour_index);
      swap_global(&f10[this_index],&f1[neighbour_index],&_max);
    }
  }

  // Swap 2 and 11
  neighbour_coord = (int4)(get_next_x(x,nx,2,periodicX),
                            get_next_y(y,ny,2,periodicY),
                            get_next_z(z,nz,2,periodicZ),
                            0);
  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f11 in %d with f2 in %d \n",this_index, neighbour_index);
      swap_global(&f11[this_index],&f2[neighbour_index],&_max);
    }
  }
  // Swap 3 and 12
  neighbour_coord = (int4)(get_next_x(x,nx,3,periodicX),
                            get_next_y(y,ny,3,periodicY),
                            get_next_z(z,nz,3,periodicZ),
                            0);
  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f12 in %d with f3 in %d \n",this_index, neighbour_index);
      swap_global(&f12[this_index],&f3[neighbour_index],&_max);
    }
  }

    // Swap 4 and 13
  neighbour_coord = (int4)(get_next_x(x,nx,4,periodicX),
                            get_next_y(y,ny,4,periodicY),
                            get_next_z(z,nz,4,periodicZ),
                            0);

  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f13 in %d with f4 in %d \n",this_index, neighbour_index);
      swap_global(&f13[this_index],&f4[neighbour_index],&_max);
    }
  }
    // Swap 5 and 14
  neighbour_coord = (int4)(get_next_x(x,nx,5,periodicX),
                            get_next_y(y,ny,5,periodicY),
                            get_next_z(z,nz,5,periodicZ),
                            0);

  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;

    if(neighbour_index >= 0){
      //printf("swaping f14 in %d with f5 in %d \n",this_index, neighbour_index);
      swap_global(&f14[this_index],&f5[neighbour_index],&_max);
    }
  }

  // Swap 6 and 15
  neighbour_coord = (int4)(get_next_x(x,nx,6,periodicX),
                            get_next_y(y,ny,6,periodicY),
                            get_next_z(z,nz,6,periodicZ),
                            0);

  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f15 in %d with f6 in %d \n",this_index, neighbour_index);
      swap_global(&f15[this_index],&f6[neighbour_index],&_max);
    }
  }
  // Swap 7 and 16
  neighbour_coord = (int4)(get_next_x(x,nx,7,periodicX),
                            get_next_y(y,ny,7,periodicY),
                            get_next_z(z,nz,7,periodicZ),
                            0);

  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f16 in %d with f7 in %d \n",this_index, neighbour_index);
      swap_global(&f16[this_index],&f7[neighbour_index],&_max);
    }
  }

  // Swap 8 and 17
  neighbour_coord = (int4)(get_next_x(x,nx,8,periodicX),
                            get_next_y(y,ny,8,periodicY),
                            get_next_z(z,nz,8,periodicZ),
                            0);

  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f17 in %d with f8 in %d \n",this_index, neighbour_index);
      swap_global(&f17[this_index],&f8[neighbour_index],&_max);
    }
  }

  // Swap 9 and 18
  neighbour_coord = (int4)(get_next_x(x,nx,9,periodicX),
                            get_next_y(y,ny,9,periodicY),
                            get_next_z(z,nz,9,periodicZ),
                            0);

  if(neighbour_coord.z >= 0 && neighbour_coord.z < nz){
    neighbour_index = read_imagei(solids,sampler,neighbour_coord).x;
    if(neighbour_index >= 0){
      //printf("swaping f18 in %d with f9 in %d \n",this_index, neighbour_index);
      swap_global(&f18[this_index],&f9[neighbour_index],&_max);
    }
  }

  max_change[this_index] = _max;

 /* 
  const int global_index = this_index;
  printf("(%d,%d,%d) ",x,y,z);
  printf("%e ",f0[global_index]);
  printf("%e ",f10[global_index]);
  printf("%e ",f11[global_index]);
  printf("%e ",f12[global_index]);
  printf("%e ",f13[global_index]);
  printf("%e ",f14[global_index]);
  printf("%e ",f15[global_index]);
  printf("%e ",f16[global_index]);
  printf("%e ",f17[global_index]);
  printf("%e ",f18[global_index]);
  printf("%e ",f1[global_index]);
  printf("%e ",f2[global_index]);
  printf("%e ",f3[global_index]);
  printf("%e ",f4[global_index]);
  printf("%e ",f5[global_index]);
  printf("%e ",f6[global_index]);
  printf("%e ",f7[global_index]);
  printf("%e ",f8[global_index]);
  printf("%e ",f9[global_index]);
  printf("\n");
*/
}

__kernel void avrage_velocity_ux(__read_only image3d_t solids,
                                __global real_t *f0,
                                __global real_t *f1,
                                __global real_t *f2,
                                __global real_t *f3,
                                __global real_t *f4,
                                __global real_t *f5,
                                __global real_t *f6,
                                __global real_t *f7,
                                __global real_t *f8,
                                __global real_t *f9,
                                __global real_t *f10,
                                __global real_t *f11,
                                __global real_t *f12,
                                __global real_t *f13,
                                __global real_t *f14,
                                __global real_t *f15,
                                __global real_t *f16,
                                __global real_t *f17,
                                __global real_t *f18,
                                __local real_t *lattice,
                                __global real_t *avrage_velocity,
                                real_t avrage_dencity,
                                int ant_vec,
                                int ghost_layers_north,
                                int ghost_layers_south,
                                int half_extra)
{


  const int x = get_global_id(0);
	const int y = get_global_id(1);
	const int z = get_global_id(2) + ghost_layers_north;

  const int nx = get_global_size(0);
	const int ny = get_global_size(1);
	const int nz = get_global_size(2);

  //const int lx = get_local_id(0);
  const int lnx = get_local_size(0);
  const int lny = get_local_size(1);
  const int lnz = get_local_size(2);

	const int lx = (get_local_id(2) * lny * lnx) + (get_local_id(1) * lnx) +  get_local_id(0);


  real_t ux = 0.0f; 

  int4 this_coord = (int4) (x,y,z,0);
  int this_index = read_imagei(solids,sampler,this_coord).x;

  //int global_index = (nx * ny * z) + (nx * y) + (x);
  const int local_index = lx * ant_vec;

  //Make pointer to this threads starting pos in the shared lattice memory
  __local real_t *local_lattice = &lattice[local_index];

  if (this_index >= 0) {   
    copy_to_shared(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,
                   f11,f12,f13,f14,f15,f16,f17,f18,
                   local_lattice,this_index);
  }


  barrier(CLK_LOCAL_MEM_FENCE);


  
  //if (z < ghost_layers_north || z >= (nz - ghost_layers_south) ){
  //  return;
  //}

  
  if (x < half_extra || x >= (nx - half_extra) ){
    avrage_velocity[this_index] = ux;
    return;
  }
 

  if (this_index >= 0) {   
    // Compute density from the particle distribution functions 
    real_t delta_rho = get_rho_shared(local_lattice);
	  real_t rho = delta_rho + avrage_dencity;

    //Compute velocity in x direction from the particle distribution functions
    ux = get_ux_shared(local_lattice,rho);
    avrage_velocity[this_index] = ux;
   
  }

   //avrage_velocity[this_index] = ux;


}

__kernel void reduce(__global real_t* buffer,
            __local real_t* scratch,
            __const int length,
            __global real_t* result) {
  
  int global_index = get_global_id(0);
  int local_index = get_local_id(0);

  //printf("gid: %d lsiz: %d\n", global_index, get_local_size(0));
  
  /*
  real_t accum[2] = {0,0};
  int counter = 0;
  // Loop sequentially over chunks of input vector
  while (global_index < length) {
    real_t element = buffer[global_index];
    accum[counter&1] += element;
    global_index += get_global_size(0);
    counter++;
  }
*/
  
  real_t accumulator = 0;
  while (global_index < length) {
    //printf("%d is adding index %d: %f to the acc %f\n",get_global_id(0),global_index,buffer[global_index],accumulator);
    real_t element = buffer[global_index];
    accumulator += element;
    global_index += get_global_size(0);
  }
  
  // Perform parallel reduction
  
  //scratch[local_index] = accum[0] + accum[1]; //accumulator +accumulator2;
  scratch[local_index] = accumulator;
  barrier(CLK_LOCAL_MEM_FENCE);


  //printf("scratch %d = %f\n",local_index,scratch[local_index]);
  for(int offset = get_local_size(0) / 2;
      offset > 0;
      offset = offset / 2) {
    if (local_index < offset) {
      real_t other = scratch[local_index + offset];
      scratch[local_index] += other;
    }
    barrier(CLK_LOCAL_MEM_FENCE);
    //printf("scratch %d = %f\n",local_index,scratch[local_index]);
  }
  if (local_index == 0) {
    result[get_group_id(0)] = scratch[0];
    //printf("result GPU = %e\n",result[get_group_id(0)]);
  }
}

__kernel void copy_to_from_ghost_layer(__global real_t *f0,
                                __global real_t *f1,
                                __global real_t *f2,
                                __global real_t *f3,
                                __global real_t *f4,
                                __global real_t *f5,
                                __global real_t *f6,
                                __global real_t *f7,
                                __global real_t *f8,
                                __global real_t *f9,
                                __global real_t *f10,
                                __global real_t *f11,
                                __global real_t *f12,
                                __global real_t *f13,
                                __global real_t *f14,
                                __global real_t *f15,
                                __global real_t *f16,
                                __global real_t *f17,
                                __global real_t *f18,
                                __global real_t *buffer,
                                int ant_vec,
                                int start_index,
                                int to)
{

  const int offset = get_global_id(0);
  const int ant_element = get_global_size(0);
  const int index = start_index + offset;

  
  __global real_t *f[19];
  f[ 0] = f0;
  f[ 1] = f1;
  f[ 2] = f2;
  f[ 3] = f3;
  f[ 4] = f4;
  f[ 5] = f5;
  f[ 6] = f6;
  f[ 7] = f7;
  f[ 8] = f8;
  f[ 9] = f9;
  f[10] = f10;
  f[11] = f11;
  f[12] = f12;
  f[13] = f13;
  f[14] = f14;
  f[15] = f15;
  f[16] = f16;
  f[17] = f17;
  f[18] = f18;

  for (int i = 0 ; i <ant_vec; i++){
    if(to == 1){
      buffer[(i * ant_element) + offset] = f[i][index];
    }else{
      f[i][index] = buffer[(i * ant_element) + offset];
    }
  }
}